"""ELM import utility."""
import importlib
import logging
from warnings import warn

logger = logging.getLogger(__name__)


def try_import(package_name):
    """Function to import packages needed for
    postgres and aws.

    Parameters
    ----------
    package_name : str
        Name of package to import.

    Returns
    -------
    p : module
        imported package.
    """
    try:
        p = importlib.import_module(package_name)
        return p
    except ImportError:
        msg = (f'Unable to import {package_name}. '
               'Please ensure you have the package '
               'installed. This is an extra requirement '
               'for the package you are running')
        logger.warning(msg)
        warn(msg)



# -*- coding: utf-8 -*-
"""
Utilities for retrieving data from OSTI.
"""
import copy
import requests
import json
import os
import pandas as pd
import logging


logger = logging.getLogger(__name__)


class OstiRecord(dict):
    """Class to handle a single OSTI record as dictionary data"""

    def __init__(self, record):
        """
        Parameters
        ----------
        record : dict
            OSTI record in dict form, typically a response from OSTI API.
        """
        assert isinstance(record, dict)
        super().__init__(**record)

    @staticmethod
    def strip_nested_brackets(text):
        """Remove text between brackets/parentheses for cleaning OSTI text"""
        ret = ''
        skip1c = 0
        skip2c = 0
        for i in text:
            if i == '[':
                skip1c += 1
            elif i == '(':
                skip2c += 1
            elif i == ']' and skip1c > 0:
                skip1c -= 1
            elif i == ')' and skip2c > 0:
                skip2c -= 1
            elif skip1c == 0 and skip2c == 0:
                ret += i
        return ret

    @property
    def authors(self):
        """Get the list of authors of this record.

        Returns
        -------
        str
        """
        au = copy.deepcopy(self.get('authors', None))
        if au is not None:
            for i, name in enumerate(au):
                name = self.strip_nested_brackets(name)
                if name.count(',') == 1:
                    second, first = name.split(',')
                    name = f'{first.strip()} {second.strip()}'
                au[i] = name
            au = ', '.join(au)
        return au

    @property
    def title(self):
        """Get the title of this record

        Returns
        -------
        str | None
        """
        return self.get('title', None)

    @property
    def year(self):
        """Get the year of publication of this record

        Returns
        -------
        str | None
        """
        year = self.get('publication_date', None)
        if year is not None:
            year = year.split('-')[0]
            year = str(year)
        return year

    @property
    def date(self):
        """Get the date of publication of this record

        Returns
        -------
        str | None
        """
        date = self.get('publication_date', None)
        if date is not None:
            date = date.split('T')[0]
            date = str(date)
        return date

    @property
    def doi(self):
        """Get the DOI of this record

        Returns
        -------
        str | None
        """
        return self.get('doi', None)

    @property
    def osti_id(self):
        """Get the OSTI ID of this record which is typically a 7 digit number

        Returns
        -------
        str | None
        """
        return self.get('osti_id', None)

    @property
    def url(self):
        """Get the download URL of this record

        Returns
        -------
        str | None
        """
        url = None
        for link in self['links']:
            if link.get('rel', None) == 'fulltext':
                url = link.get('href', None)
                break
        return url

    def download(self, fp):
        """Download the PDF of this record

        Parameters
        ----------
        fp : str
            Filepath to download this record to, typically a .pdf
        """
        # OSTI returns citation on first query and pdf on second (weird)
        session = requests.Session()
        response = session.get(self.url)
        response = session.get(self.url)
        with open(fp, 'wb') as f_pdf:
            f_pdf.write(response.content)


class OstiList(list):
    """Class to retrieve and handle multiple OSTI records from an API URL."""

    BASE_URL = 'https://www.osti.gov/api/v1/records'
    """Base OSTI API URL. This can be appended with search parameters"""

    def __init__(self, url, n_pages=1):
        """
        Parameters
        ----------
        url : str
            OSTI API URL to request, see this for details:
                https://www.osti.gov/api/v1/docs
        n_pages : int
            Number of pages to get from the API. Typical response has 20
            entries per page. Default of 1 ensures that this class doesnt hang
            on a million responses.
        """

        self.url = url
        self._session = requests.Session()
        self._response = None
        self._n_pages = 0
        self._iter = 0

        records = self._get_all(n_pages)
        records = [OstiRecord(single) for single in records]
        super().__init__(records)

    def _get_first(self):
        """Get the first page of OSTI records

        Returns
        -------
        list
        """
        self._response = self._session.get(self.url)

        if not self._response.ok:
            msg = ('OSTI API Request got error {}: "{}"'
                   .format(self._response.status_code,
                           self._response.reason))
            raise RuntimeError(msg)
        first_page = self._response.json()

        self._n_pages = 1
        if 'last' in self._response.links:
            url = self._response.links['last']['url']
            self._n_pages = int(url.split('page=')[-1])

        logger.debug('Found approximately {} records.'
                     .format(self._n_pages * len(first_page)))

        return first_page

    def _get_pages(self, n_pages):
        """Get response pages up to n_pages from OSTI.

        Parameters
        ----------
        n_pages : int
            Number of pages to retrieve

        Returns
        -------
        next_pages : list
            This function will return a generator of next pages, each of which
            is a list of OSTI records
        """
        if n_pages > 1:
            for page in range(2, self._n_pages + 1):
                if page <= n_pages:
                    next_page = self._session.get(self.url,
                                                  params={'page': page})
                    next_page = next_page.json()
                    yield next_page
                else:
                    break

    def _get_all(self, n_pages):
        """Get all pages of records up to n_pages.

        Parameters
        ----------
        n_pages : int
            Number of pages to retrieve

        Returns
        -------
        all_records : list
            List of all records.
        """
        first_page = self._get_first()
        records = first_page

        for page in self._get_pages(n_pages):
            records.extend(page)

        return records

    def download(self, out_dir):
        """Download all PDFs from the records in this OSTI object into a
        directory. PDFs will be given file names based on their OSTI record ID

        Parameters
        ----------
        out_dir : str
            Directory to download PDFs to. This directory will be created if it
            does not already exist.
        """
        logger.info('Downloading {} records to: {}'.format(len(self), out_dir))
        os.makedirs(out_dir, exist_ok=True)
        for record in self:
            fp_out = os.path.join(out_dir, record.osti_id + '.pdf')
            if not os.path.exists(fp_out):
                try:
                    record.download(fp_out)
                except Exception as e:
                    logger.exception('Could not download OSTI ID {} "{}": {}'
                                     .format(record.osti_id, record.title, e))
        logger.info('Finished download!')

    @property
    def meta(self):
        """Get a meta dataframe with details on all of the OSTI records.

        Returns
        -------
        pd.DataFrame
        """
        i = 0
        attrs = ('authors', 'title', 'year', 'date', 'doi', 'osti_id', 'url')
        df = pd.DataFrame(columns=attrs)
        for record in self:
            for attr in attrs:
                out = getattr(record, attr)
                if not isinstance(out, str):
                    out = json.dumps(out)
                df.at[i, attr] = out
            df.at[i, 'fn'] = f'{record.osti_id}.pdf'
            i += 1
        return df

    @classmethod
    def from_osti_ids(cls, oids):
        """Initialize OSTI records from one or more numerical IDS

        Parameters
        ----------
        oids : list
            List of string or integer OSTI IDs which are typically 7 digit
            numbers

        Returns
        -------
        out : OstiList
            OstiList object with entries for each oid input.
        """
        if not isinstance(oids, (list, tuple)):
            oids = [oids]
        oids = [str(oid) for oid in oids]
        out = None
        for oid in oids:
            iout = cls(cls.BASE_URL + '/' + oid)
            if out is None:
                out = iout
            else:
                out += iout
        return out



# -*- coding: utf-8 -*-
"""ELM Web Scraping - Google search."""
import asyncio
import logging

from playwright.async_api import (
    async_playwright,
    TimeoutError as PlaywrightTimeoutError,
)

from elm.web.utilities import clean_search_query


logger = logging.getLogger(__name__)
_SEARCH_RESULT_TAG = '[jsname="UWckNb"]'


class PlaywrightGoogleLinkSearch:
    """Search for top results on google and return their links"""

    EXPECTED_RESULTS_PER_PAGE = 10
    """Number of results displayed per Google page. """

    def __init__(self, **launch_kwargs):
        """

        Parameters
        ----------
        **launch_kwargs
            Keyword arguments to be passed to
            `playwright.chromium.launch`. For example, you can pass
            ``headless=False, slow_mo=50`` for a visualization of the
            search.
        """
        self.launch_kwargs = launch_kwargs
        self._browser = None

    async def _load_browser(self, pw_instance):
        """Launch a chromium instance and load a page"""
        self._browser = await pw_instance.chromium.launch(**self.launch_kwargs)

    async def _close_browser(self):
        """Close browser instance and reset internal attributes"""
        await self._browser.close()
        self._browser = None

    async def _search(self, query, num_results=10):
        """Search google for links related to a query."""
        logger.debug("Searching Google: %r", query)
        num_results = min(num_results, self.EXPECTED_RESULTS_PER_PAGE)

        page = await self._browser.new_page()
        await _navigate_to_google(page)
        await _perform_google_search(page, query)
        return await _extract_links(page, num_results)

    async def _skip_exc_search(self, query, num_results=10):
        """Perform search while ignoring timeout errors"""
        try:
            return await self._search(query, num_results=num_results)
        except PlaywrightTimeoutError as e:
            logger.exception(e)
            return []

    async def _get_links(self, queries, num_results):
        """Get links for multiple queries"""
        outer_task_name = asyncio.current_task().get_name()
        async with async_playwright() as pw_instance:
            await self._load_browser(pw_instance)
            searches = [
                asyncio.create_task(
                    self._skip_exc_search(query, num_results=num_results),
                    name=outer_task_name,
                )
                for query in queries
            ]
            results = await asyncio.gather(*searches)
            await self._close_browser()
        return results

    async def results(self, *queries, num_results=10):
        """Retrieve links for the first `num_results` of each query.

        This function executes a google search for each input query and
        returns a list of links corresponding to the top `num_results`.

        Parameters
        ----------
        num_results : int, optional
            Number of top results to retrieve for each query. Note that
            this value can never exceed the number of results per page
            (typically 10). If you pass in a larger value, it will be
            reduced to the number of results per page.
            By default, ``10``.

        Returns
        -------
        list
            List equal to the length of the input queries, where each
            entry is another list containing the top `num_results`
            links.
        """
        queries = map(clean_search_query, queries)
        return await self._get_links(queries, num_results)


async def _navigate_to_google(page):
    """Navigate to Google domain."""
    await page.goto("https://www.google.com")
    await page.wait_for_load_state("networkidle")


async def _perform_google_search(page, search_query):
    """Fill in search bar with user query and click search button"""
    await page.get_by_label("Search", exact=True).fill(search_query)
    await _close_autofill_suggestions(page)
    await page.get_by_role("button", name="Google Search").click()


async def _close_autofill_suggestions(page):
    """Google autofill suggestions often get in way of search button.

    We get around this by closing the suggestion dropdown before
    looking for the search button. Looking for the "Google Search"
    button doesn't work because it is sometimes obscured by the dropdown
    menu. Clicking the "Google" logo can also fail when they add
    seasonal links/images (e.g. holiday logos). Current solutions is to
    look for a specific div at the top of the page.
    """
    await page.locator("#gb").click()


async def _extract_links(page, num_results):
    """Extract links for top `num_results` on page"""
    links = await asyncio.to_thread(page.locator, _SEARCH_RESULT_TAG)
    return [
        await links.nth(i).get_attribute("href") for i in range(num_results)
    ]



# -*- coding: utf-8 -*-
"""
Research Summarization and Distillation with LLMs
"""
import logging
import os

from elm.base import ApiBase
from elm.chunk import Chunker


logger = logging.getLogger(__name__)


class Summary(ApiBase):
    """Interface to perform Recursive Summarization and Distillation of
    research text"""

    MODEL_ROLE = "You are an energy scientist summarizing prior research"
    """High level model role, somewhat redundant to MODEL_INSTRUCTION"""

    MODEL_INSTRUCTION = ('Can you please summarize the text quoted above '
                         'in {n_words} words?\n\n"""\n{text_chunk}\n"""')
    """Prefix to the engineered prompt. The format args `text_chunk` and
    `n_words` will be formatted by the Summary class at runtime. `text_chunk`
    will be provided by the Summary text chunks, `n_words` is an initialization
    argument for the Summary class."""

    def __init__(self, text, model=None, n_words=500, **chunk_kwargs):
        """
        Parameters
        ----------
        text : str | list
            Single body of text to chunk up using elm.Chunker or a pre-chunked
            list of strings. Works well if this is a single document with empty
            lines between paragraphs.
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        n_words : int
            Desired length of the output text. Note that this is never perfect
            but helps guide the LLM to an approximate desired output length.
            400-600 words seems to work quite well with GPT-4. This gets
            formatted into the MODEL_INSTRUCTION attribute.
        chunk_kwargs : dict | None
            kwargs for initialization of :class:`elm.chunk.Chunker`
        """

        super().__init__(model)

        self.text = text
        self.n_words = n_words

        assert isinstance(self.text, (str, list, tuple))

        if isinstance(self.text, str):
            if os.path.isfile(text):
                logger.info('Loading text file: {}'.format(text))
                with open(text, 'r') as f:
                    self.text = f.read()
            self.text_chunks = Chunker(self.text, **chunk_kwargs)
        else:
            self.text_chunks = self.text

        self.summary_chunks = []

    def combine(self, text_summary):
        """Combine separate chunk summaries into one more comprehensive
        narrative

        Parameters
        ----------
        summary : str
            Summary of text. May be several disjointed paragraphs

        Returns
        -------
        summary : str
            Summary of text. Paragraphs will be more cohesive.
        """
        role = 'You provide editorial services for technical writing.'
        query = ('Can you combine the following paragraphs and '
                 'ease the transitions between them? '
                 f'\n\n"""{text_summary}"""')
        text_summary = self.generic_query(query, model_role=role)
        return text_summary

    def run(self, temperature=0, fancy_combine=True):
        """Use GPT to do a summary of input text.

        Parameters
        ----------
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.
        fancy_combine : bool
            Flag to use the GPT model to combine the separate outputs into a
            cohesive summary.

        Returns
        -------
        summary : str
            Summary of text.
        """

        logger.info('Summarizing {} text chunks in serial...'
                    .format(len(self.text_chunks)))
        summary = ''

        for i, chunk in enumerate(self.text_chunks):
            logger.debug('Summarizing text chunk {} out of {}'
                         .format(i + 1, len(self.text_chunks)))

            msg = self.MODEL_INSTRUCTION.format(text_chunk=chunk,
                                                n_words=self.n_words)
            response = self.generic_query(msg, model_role=self.MODEL_ROLE,
                                          temperature=temperature)
            self.summary_chunks.append(response)
            summary += f'\n\n{response}'

        if fancy_combine:
            summary = self.combine(summary)

        logger.info('Finished all summaries.')

        return summary

    async def run_async(self, temperature=0, ignore_error=None,
                        rate_limit=40e3, fancy_combine=True):
        """Run text summary asynchronously for all text chunks

        NOTE: you need to call this using the await command in ipython or
        jupyter, e.g.: `out = await Summary.run_async()`

        Parameters
        ----------
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.
        ignore_error : None | callable
            Optional callable to parse API error string. If the callable
            returns True, the error will be ignored, the API call will not be
            tried again, and the output will be an empty string.
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            gpt-3.5-turbo limit is 90k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.
        fancy_combine : bool
            Flag to use the GPT model to combine the separate outputs into a
            cohesive summary.

        Returns
        -------
        summary : str
            Summary of text.
        """

        logger.info('Summarizing {} text chunks asynchronously...'
                    .format(len(self.text_chunks)))

        queries = []
        for chunk in self.text_chunks:
            msg = self.MODEL_INSTRUCTION.format(text_chunk=chunk,
                                                n_words=self.n_words)
            queries.append(msg)

        summaries = await self.generic_async_query(queries,
                                                   model_role=self.MODEL_ROLE,
                                                   temperature=temperature,
                                                   ignore_error=ignore_error,
                                                   rate_limit=rate_limit)

        self.summary_chunks = summaries
        summary = '\n\n'.join(summaries)

        if fancy_combine:
            summary = self.combine(summary)

        logger.info('Finished all summaries.')

        return summary



"""
Code to build Corpus from the researcher hub.
"""
import os
import os.path
import logging
import json
import math
import re
import requests
import pandas as pd

logger = logging.getLogger(__name__)


class ProfilesRecord(dict):
    """Class to handle a single profiles as dictionary data.
    This class requires setting an 'RHUB_API_KEY' environment
    variable to access the Pure Web Service. The API key can be
    obtained by contacting an NREL library representative:
    Library@nrel.gov.
    """
    def __init__(self, record):
        """
        Parameters
        ----------
        record : dict
            Profile in dict form, typically a response from the API.
        """
        api_key = os.getenv("RHUB_API_KEY")
        assert api_key is not None, "Please set RHUB_API_KEY!"
        assert isinstance(record, dict)

        super().__init__(**record)

    @staticmethod
    def clean_text(html_text):
        """Clean html text from API response

        Parameters
        ----------
        html_text : str
            Text containing html characters.
        Returns
        -------
        clean : str
            Text with html characters removed.
        """
        clean = re.sub(r'<.*?>', '', html_text)
        clean = clean.replace('\xa0', ' ')

        return clean

    @property
    def first_name(self):
        """Get the first name of this researcher.

        Returns
        -------
        first : str
            Full name of researcher.
        """
        names = self.get('name')
        first = names.get('firstName')

        return first

    @property
    def last_name(self):
        """Get the last name of this researcher.

        Returns
        -------
        last : str
            Last name of researcher.
        """
        names = self.get('name')
        last = names.get('lastName')

        return last

    @property
    def title(self):
        """Get the full name of this researcher.

        Returns
        -------
        full : str
            Full name of researcher.
        """
        names = self.get('name')
        first = names.get('firstName')
        last = names.get('lastName')
        full = first + ' ' + last

        return full

    @property
    def email(self):
        """Get the email address of this researcher.

        Returns
        -------
        email : str
            Email address of researcher.
        """
        email = None
        orgs = self.get('staffOrganisationAssociations')
        if orgs:
            emails_dict = orgs[0].get('emails')
            if emails_dict:
                email = emails_dict[0].get('value').get('value')

        return email

    @property
    def url(self):
        """Get the url or this researcher's profile.

        Returns
        -------
        url : str
            URL to researcher's profile.
        """
        info = self.get('info')
        url = info.get('portalUrl')

        return url

    @property
    def id(self):
        """Get API ID of researcher.

        Returns
        -------
        id : str
            Researcher ID.
        """
        level = self.get('ids')[0]
        id = level.get('value').get('value')

        return id

    @property
    def position(self):
        """Get the position of this researcher.

        Returns
        -------
        position : str
            Researcher's position.
        """
        position = None
        org = self.get('staffOrganisationAssociations')
        if org:
            info = org[0].get('jobDescription')
            text = info.get('text')[0]
            position = text.get('value')

        return position

    @property
    def profile_information(self):
        """Get key profile information for this record:
        Personal Profile, Research Interests, Professional Experience.

        Returns
        -------
        bio : str
            Researcher's profile text.
        interests : str
            Text from Research Interests section.
        experience : str
            Text from Professional Experience section.
        """
        prof = self.get('profileInformations')

        bio = None
        interests = None
        experience = None

        if prof:
            for section in prof:
                type = section.get('type').get('term')
                if 'Personal Profile' in str(type):
                    info = section.get('value').get('text')[0]
                    bio = info.get('value')
                    bio = self.clean_text(bio)

                if 'Research Interests' in str(type):
                    info = section.get('value').get('text')[0]
                    interests = info.get('value')
                    interests = self.clean_text(interests)

                if 'Professional Experience' in str(type):
                    info = section.get('value').get('text')[0]
                    experience = info.get('value')
                    experience = self.clean_text(experience)

        return bio, interests, experience

    @property
    def education(self):
        """Get the education information of this researcher.

        Returns
        -------
        levels : list
            Degree levels, ex: Master, Bachelor, PhD
        degs : list
            Area of study, ex: Mechanical Engineering
        schools : list
            School awarding the degree.
        """
        researcher_name = self.title
        edu = self.get('educations')
        out_strings = []

        if edu:
            for e in edu:
                try:
                    if e.get('projectTitle'):
                        quali = e.get('qualification')
                        level = quali.get('term').get('text')[0].get('value')
                        deg = e.get('projectTitle').get('text')[0].get('value')
                        org = e.get('organisationalUnits')

                        if org:
                            value = org[0].get('externalOrganisationalUnit')
                            name = value.get('name')
                            school = name.get('text')[0].get('value')
                        else:
                            deg_school = deg
                            deg = deg_school.split(',')[0]
                            school = deg_school.split(',')[1]

                        deg_string = (f'{researcher_name} has a {level} '
                                      f'degree in {deg} from {school}. ')
                        out_strings.append(deg_string)
                    else:
                        quali = e.get('qualification')
                        level = quali.get('term').get('text')[0].get('value')
                        org = e.get('organisationalUnits')[0]
                        org_unit = org.get('externalOrganisationalUnit')
                        name = org_unit.get('name')
                        school = name.get('text')[0].get('value')

                        deg_string = (f'{researcher_name} has a {level} '
                                      f'degree from {school}. ')
                        out_strings.append(deg_string)
                except Exception:
                    pass

            return out_strings

    @property
    def publications(self):
        """Get the publications this researcher contributed to.

        Returns
        -------
        pubs : list
            All publications associated with this researcher.
        """
        api_key = os.getenv("RHUB_API_KEY")
        assert api_key is not None, "Please set RHUB_API_KEY!"

        id = self.get('pureId')
        url = (f'https://research-hub.nrel.gov/ws/api/524/persons/'
               f'{id}/research-outputs?size=100'
               f'&apiKey={api_key}')
        session = requests.Session()
        response = session.get(url, headers={'Accept': 'application/json'})

        content = response.json()['items']

        pubs = []

        for pub in content:
            title = pub.get('title').get('value')
            pubs.append(title)

        return pubs

    def download(self, fp):
        """Download text file containing researchers profile information.

        Parameters
        ----------
        fp : str
            Filepath to download this record to.
        """
        name = self.title

        if self.position:
            full = (f"The following is a brief biography for {name} "
                    f"who is a {self.position} for the National Renewable "
                    f"Energy Laboratory: ")
        else:
            full = (f"The following is a brief biography for {name} "
                    f"who works for the National Renewable "
                    f"Energy Laboratory: ")

        profile, interests, experience = self.profile_information

        if profile:
            full += profile + ' '

        if interests:
            research = (f"{name}'s research interests include: "
                        f"{interests}. ")
            full += research

        if experience:
            research = (f"{name}'s professional experience includes: "
                        f"{experience}. ")
            full += research

        if self.education:
            for edu in self.education:
                full += edu

        if self.publications:
            publications = (f"{name} has been involved in the following "
                            f"publications: {', '.join(self.publications)}. ")
            full += publications

        with open(fp, "w") as text_file:
            text_file.write(full)


class ProfilesList(list):
    """Class to retrieve and handle multiple profiles from an API URL.
    This class requires setting an 'RHUB_API_KEY' environment
    variable to access the Pure Web Service. The API key can be
    obtained by contacting an NREL library representative:
    Library@nrel.gov.
    """
    def __init__(self, url, n_pages=1):
        """
        Parameters
        ----------
        url : str
            Research Hub API URL to request, see this for details:
                https://research-hub.nrel.gov/ws/api/524/api-docs/index.html
        n_pages : int
            Number of pages to get from the API. Typical response has 20
            entries per page. Default of 1 ensures that this class doesnt hang
            on a million responses.
        """
        api_key = os.getenv("RHUB_API_KEY")
        assert api_key is not None, "Please set RHUB_API_KEY!"

        self.url = url
        self._session = requests.Session()
        self._response = None
        self._n_pages = 0
        self._iter = 0

        records = self._get_all(n_pages)
        records = [ProfilesRecord(single) for single in records]
        records = [prof for prof in records if prof.last_name != 'NREL']
        super().__init__(records)

    def _get_first(self):
        """Get the first page of Profiles.

        Returns
        -------
        first_page : list
            First page of records as a list.
        """
        self._response = self._session.get(
            self.url,
            headers={'Accept': 'application/json'}
        )

        resp = self._response.json()

        if not self._response.ok:
            msg = ('API Request got error {}: "{}"'
                   .format(self._response.status_code,
                           self._response.reason))
            raise RuntimeError(msg)
        first_page = self._response.json()['items']

        self._n_pages = 1

        if 'last' not in self._response.links:
            count_pages = resp['count'] / resp['pageInformation'].get('size')
            self._n_pages = math.ceil(count_pages)
        else:
            url = self._response.links['last']['url']
            self._n_pages = int(url.split('page=')[-1])

        logger.debug('Found approximately {} records.'
                     .format(self._n_pages * len(first_page)))

        return first_page

    def _get_pages(self, n_pages):
        """Get response pages up to n_pages from Research Hub.

        Parameters
        ----------
        n_pages : int
            Number of pages to retrieve

        Returns
        -------
        next_pages : list
            This function will return a generator of next pages, each of which
            is a list of profiles.
        """
        if n_pages > 1:
            for page in range(2, self._n_pages + 1):
                if page <= n_pages:
                    next_page = self._session.get(
                        self.url,
                        params={'page': page},
                        headers={'Accept': 'application/json'})

                    next_page = next_page.json()['items']
                    yield next_page
                else:
                    break

    def _get_all(self, n_pages):
        """Get all pages of profiles up to n_pages.

        Parameters
        ----------
        n_pages : int
            Number of pages to retrieve

        Returns
        -------
        all_records : list
            List of all publication records.
        """
        first_page = self._get_first()
        records = first_page

        for page in self._get_pages(n_pages):
            records.extend(page)

        return records

    def meta(self):
        """Get a meta dataframe with details on all of the profiles.

        Returns
        -------
        df : pd.DataFrame
            Dataframe containing all metadata information.
        """
        i = 0
        attrs = ('title', 'email', 'url', 'id')
        df = pd.DataFrame(columns=attrs)
        for record in self:
            for attr in attrs:
                out = getattr(record, attr)
                if not isinstance(out, str):
                    out = json.dumps(out)
                df.at[i, attr] = out
            df.at[i, 'fn'] = f'{record.id}.txt'
            df.at[i, 'category'] = 'Researcher Profile'
            i += 1

        return df

    def download(self, out_dir):
        """Download all profiles from the records in this object into a
        directory. TXT files will be given file names based on researcher ID.

        Parameters
        ----------
        out_dir : str
            Directory to download TXT files to. This directory will be created
            if it does not already exist.
        """
        os.makedirs(out_dir, exist_ok=True)
        for record in self:
            fn = record.id
            fp_out = os.path.join(out_dir, fn + '.txt')
            if not os.path.exists(fp_out):
                try:
                    record.download(fp_out)
                except Exception as e:
                    print(f"Could not download {record.title} with error {e}")
                    logger.exception('Could not download profile ID {}: {}'
                                     .format(record.title, e))
        logger.info('Finished Profiles download!')


class PublicationsRecord(dict):
    """Class to handle a single publication as dictionary data.
    This class requires setting an 'RHUB_API_KEY' environment
    variable to access the Pure Web Service. The API key can be
    obtained by contacting an NREL library representative:
    Library@nrel.gov.
    """
    def __init__(self, record):
        """
        Parameters
        ----------
        record : dict
            Research Hub record in dict form, typically a response from API.
        """
        api_key = os.getenv("RHUB_API_KEY")
        assert api_key is not None, "Please set RHUB_API_KEY!"

        assert isinstance(record, dict)
        super().__init__(**record)

    @property
    def title(self):
        """Get the title of this publication.

        Returns
        -------
        title : str
            Publication title.
        """
        title = self.get('title').get('value')

        return title

    @property
    def year(self):
        """Get the publish year.

        Returns
        -------
        year : int
            Year of publication.
        """
        status = self.get('publicationStatuses')[0]
        year = status.get('publicationDate').get('year')

        return year

    @property
    def url(self):
        """Get the url associated with the publication.

        Returns
        -------
        url : str
            Publication URL.
        """
        info = self.get('info')
        url = info.get('portalUrl')

        return url

    @property
    def id(self):
        """Get the 'NREL Publication Number' for
        this record.

        Returns
        -------
        id : str
            Publication Number.
        """
        try:
            group = self.get('keywordGroups')[0]
            cont = group.get('keywordContainers')[0]
            id = cont.get('freeKeywords')[0].get('freeKeywords')[0]
            id = id.replace('/', '-')
        except TypeError:
            id = self.get('externalId')

        return id

    @property
    def authors(self):
        """Get the names of all authors for a publication.

        Returns
        -------
        out : str
            String containing author names.
        """
        pa = self.get('personAssociations')

        authors = []

        for r in pa:
            first = r.get('name').get('firstName')
            last = r.get('name').get('lastName')

            if first and last:
                full = first + ' ' + last
            elif first:
                full = first
            elif last:
                full = last

            authors.append(full)

            out = ', '.join(authors)

        return out

    @property
    def category(self):
        """Get the publication category for this record.

        Returns
        -------
        cat : str
            Publication category, ex: Technical Report, Article.
        """
        type = self.get('type')
        term = type.get('term')
        cat = term.get('text')[0].get('value')

        return cat

    @property
    def links(self):
        """Get the doi and pdf links for a publication.

        Returns
        -------
        doi : str
            doi link for publication.
        pdf_url : str
            pdf link for publication.
        """
        ev = self.get('electronicVersions')

        doi = None
        pdf_url = None
        if ev:
            for link in ev:
                if link.get('doi'):
                    doi = link.get('doi')
                if link.get('link'):
                    pdf_url = link.get('link')

        return doi, pdf_url

    @property
    def abstract(self):
        """Get the abstract text for this publication.

        Returns
        -------
        value : str
            String containing abstract text.
        """
        abstract = self.get('abstract')
        text = abstract.get('text')[0]
        value = text.get('value')

        return value

    def save_abstract(self, abstract_text, out_fp):
        """Download abstract text to .txt file to the directory
        provided.

        Parameters
        ----------
        abstract_text : str
            String with abstract text.
        out_dir : str
            Directory to download TXT files to.
        """
        title = self.title
        full = f"The report titled {title} can be summarized as follows: "
        full += abstract_text

        with open(out_fp, "w") as text_file:
            text_file.write(full)

    def download(self, pdf_dir, txt_dir):
        """Download PDFs and TXT files to the directories provided. If a record
        does not fit the criteria for PDF download, a TXT file with the record
        abstract will be saved to the TXT directory.

        Parameters
        ----------
        pdf_dir : str
            Directory for pdf download.
        txt_dir : str
            Directory for txt download.
        """

        category = self.category
        pdf_url = self.links[1]
        abstract = self.abstract

        pdf_categories = ['Technical Report', 'Paper', 'Fact Sheet']

        if category not in pdf_categories:
            fn = self.id.replace('/', '-') + '.txt'
            fp = os.path.join(txt_dir, fn)
            if not os.path.exists(fp):
                if abstract:
                    self.save_abstract(abstract, fp)
        else:
            if pdf_url and pdf_url.endswith('.pdf'):
                fn = self.id.replace('/', '-') + '.pdf'
                fp = os.path.join(pdf_dir, fn)
                if not os.path.exists(fp):
                    session = requests.Session()
                    response = session.get(pdf_url)
                    with open(fp, 'wb') as f_pdf:
                        f_pdf.write(response.content)
            else:
                fn = self.id.replace('/', '-') + '.txt'
                fp = os.path.join(txt_dir, fn)
                self.save_abstract(abstract, fp)


class PublicationsList(list):
    """Class to retrieve and handle multiple publications from an API URL.
    This class requires setting an 'RHUB_API_KEY' environment
    variable to access the Pure Web Service. The API key can be
    obtained by contacting an NREL library representative:
    Library@nrel.gov.
    """
    def __init__(self, url, n_pages=1):
        """
        Parameters
        ----------
        url : str
            Research Hub API URL to request, see this for details:
                https://research-hub.nrel.gov/ws/api/524/api-docs/index.html
        n_pages : int
            Number of pages to get from the API. Typical response has 20
            entries per page. Default of 1 ensures that this class doesnt hang
            on a million responses.
        """
        api_key = os.getenv("RHUB_API_KEY")
        assert api_key is not None, "Please set RHUB_API_KEY!"

        self.url = url
        self._session = requests.Session()
        self._response = None
        self._n_pages = 0
        self._iter = 0

        records = self._get_all(n_pages)
        records = [PublicationsRecord(single) for single in records]
        super().__init__(records)

    def _get_first(self):
        """Get the first page of publications

        Returns
        -------
        first_page : list
            Publication records as list.
        """
        self._response = self._session.get(
            self.url,
            headers={'Accept': 'application/json'})

        resp = self._response.json()

        if not self._response.ok:
            msg = ('API Request got error {}: "{}"'
                   .format(self._response.status_code,
                           self._response.reason))
            raise RuntimeError(msg)
        first_page = self._response.json()['items']

        self._n_pages = 1

        if 'last' not in self._response.links:
            count_pages = resp['count'] / resp['pageInformation'].get('size')
            self._n_pages = math.ceil(count_pages)
        else:
            url = self._response.links['last']['url']
            self._n_pages = int(url.split('page=')[-1])

        logger.debug('Found approximately {} records.'
                     .format(self._n_pages * len(first_page)))

        return first_page

    def _get_pages(self, n_pages):
        """Get response pages up to n_pages from Research Hub.

        Parameters
        ----------
        n_pages : int
            Number of pages to retrieve

        Returns
        -------
        next_pages : list
            This function will return a generator of next pages, each of which
            is a list of records.
        """
        if n_pages > 1:
            for page in range(2, self._n_pages + 1):
                if page <= n_pages:
                    next_page = self._session.get(
                        self.url,
                        params={'page': page},
                        headers={'Accept': 'application/json'})

                    next_page = next_page.json()['items']
                    yield next_page
                else:
                    break

    def _get_all(self, n_pages):
        """Get all pages of publications up to n_pages.

        Parameters
        ----------
        n_pages : int
            Number of pages to retrieve

        Returns
        -------
        all_records : list
            List of all publication records.
        """
        first_page = self._get_first()
        records = first_page

        for page in self._get_pages(n_pages):
            records.extend(page)

        return records

    def meta(self):
        """Get a meta dataframe with details on all of the publications.

        Returns
        -------
        df : pd.DataFrame
            Dataframe containing all metadata information.
        """
        i = 0
        attrs = ('title', 'year', 'url', 'id', 'category', 'authors')
        df = pd.DataFrame(columns=attrs)
        for record in self:
            doi = record.links[0]
            pdf_url = record.links[1]
            for attr in attrs:
                out = getattr(record, attr)
                if not isinstance(out, str):
                    out = json.dumps(out)
                df.at[i, attr] = out
            df.at[i, 'doi'] = doi
            df.at[i, 'pdf_url'] = pdf_url
            i += 1

        return df

    def download(self, pdf_dir, txt_dir):
        """Download all PDFs and abstract TXTs from the records in this
        objectbinto a directory. Files will be given file names based
        on their record ID.

        Parameters
        ----------
        pdf_dir : str
            Directory to download PDFs to. This directory will be created
            if it does not already exist.
        txt_dir : str
            Directory to download TXTs to. This directory will be created
            if it does not already exist.
        """
        os.makedirs(pdf_dir, exist_ok=True)
        os.makedirs(txt_dir, exist_ok=True)
        for record in self:
            try:
                record.download(pdf_dir, txt_dir)
            except Exception as e:
                print(f"Could not download {record.title} with error {e}")
                logger.exception('Could not download {}: {}'
                                 .format(record.title, e))
        logger.info('Finished publications download!')



# -*- coding: utf-8 -*-
"""
ELM energy wizard
"""
from abc import ABC, abstractmethod
import copy
import os
import json
import numpy as np

from elm.base import ApiBase
from elm.utilities.try_import import try_import


class EnergyWizardBase(ApiBase, ABC):
    """Base interface to ask OpenAI LLMs about energy research."""

    MODEL_ROLE = "You parse through articles to answer questions."
    """High level model role, somewhat redundant to MODEL_INSTRUCTION"""

    MODEL_INSTRUCTION = ('Use the information below to answer the subsequent '
                         'question. If the answer cannot be found in the '
                         'text, write "I could not find an answer."')
    """Prefix to the engineered prompt"""

    def __init__(self, model=None, token_budget=3500):
        """
        Parameters
        ----------
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        token_budget : int
            Number of tokens that can be embedded in the prompt. Note that the
            default budget for GPT-3.5-Turbo is 4096, but you want to subtract
            some tokens to account for the response budget.
        """

        super().__init__(model)
        self.token_budget = token_budget

    @abstractmethod
    def query_vector_db(self, query, limit=100):
        """Returns a list of strings and relatednesses, sorted from most
        related to least.

        Parameters
        ----------
        query : str
            Question being asked of GPT
        limit : int
            Number of top results to return.

        Returns
        -------
        strings : np.ndarray
            1D array of related strings
        score : np.ndarray
            1D array of float scores of strings
        idx : np.ndarray
            1D array of indices in the text corpus corresponding to the
            ranked strings/scores outputs.
        """

    def engineer_query(self, query, token_budget=None, new_info_threshold=0.7,
                       convo=False):
        """Engineer a query for GPT using the corpus of information

        Parameters
        ----------
        query : str
            Question being asked of GPT
        token_budget : int
            Option to override the class init token budget.
        new_info_threshold : float
            New text added to the engineered query must contain at least this
            much new information. This helps prevent (for example) the table of
            contents being added multiple times.
        convo : bool
            Flag to perform semantic search with full conversation history
            (True) or just the single query (False). Call EnergyWizard.clear()
            to reset the chat history.
        Returns
        -------
        message : str
            Engineered question to GPT including information from corpus and
            the original query
        references : list
            The list of references (strs) used in the engineered prompt is
            returned here
        """

        self.messages.append({"role": "user", "content": query})

        if convo:
            # [1:] to not include the system role in the semantic search
            query = [f"{msg['role'].upper()}: {msg['content']}"
                     for msg in self.messages[1:]]

            query = '\n\n'.join(query)

        token_budget = token_budget or self.token_budget

        strings, _, idx = self.query_vector_db(query)

        message = copy.deepcopy(self.MODEL_INSTRUCTION)
        question = f"\n\nQuestion: {query}"
        used_index = []

        for string, i in zip(strings, idx):
            next_str = (f'\n\n"""\n{string}\n"""')
            token_usage = self.count_tokens(message + next_str + question,
                                            self.model)

            new_words = set(next_str.split(' '))
            additional_info = new_words - set(message.split(' '))
            new_info_frac = len(additional_info) / len(new_words)

            if new_info_frac > new_info_threshold:
                if token_usage > token_budget:
                    break
                else:
                    message += next_str
                    used_index.append(i)

        message = message + question
        used_index = np.array(used_index)
        references = self.make_ref_list(used_index)

        return message, references, used_index

    @abstractmethod
    def make_ref_list(self, idx):
        """Make a reference list

        Parameters
        ----------
        used_index : np.ndarray
            Indices of the used text from the text corpus

        Returns
        -------
        ref_list : list
            A list of references (strs) used. Ideally, this is something like:
            ["{ref_title} ({ref_url})"]
        """

    def chat(self, query,
             debug=True,
             stream=True,
             temperature=0,
             convo=False,
             token_budget=None,
             new_info_threshold=0.7,
             print_references=False,
             return_chat_obj=False):
        """Answers a query by doing a semantic search of relevant text with
        embeddings and then sending engineered query to the LLM.

        Parameters
        ----------
        query : str
            Question being asked of EnergyWizard
        debug : bool
            Flag to return extra diagnostics on the engineered question.
        stream : bool
            Flag to print subsequent chunks of the response in a streaming
            fashion
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.
        convo : bool
            Flag to perform semantic search with full conversation history
            (True) or just the single query (False). Call EnergyWizard.clear()
            to reset the chat history.
        token_budget : int
            Option to override the class init token budget.
        new_info_threshold : float
            New text added to the engineered query must contain at least this
            much new information. This helps prevent (for example) the table of
            contents being added multiple times.
        print_references : bool
            Flag to print references if EnergyWizard is initialized with a
            valid ref_col.
        return_chat_obj : bool
            Flag to only return the ChatCompletion from OpenAI API.

        Returns
        -------
        response : str
            GPT output / answer.
        query : str
            If debug is True, the engineered query asked of GPT will also be
            returned here
        references : list
            If debug is True, the list of references (strs) used in the
            engineered prompt is returned here
        """

        out = self.engineer_query(query, token_budget=token_budget,
                                  new_info_threshold=new_info_threshold,
                                  convo=convo)
        query, references, _ = out

        messages = [{"role": "system", "content": self.MODEL_ROLE},
                    {"role": "user", "content": query}]
        response_message = ''
        kwargs = dict(model=self.model,
                      messages=messages,
                      temperature=temperature,
                      stream=stream)

        response = self._client.chat.completions.create(**kwargs)

        if return_chat_obj:
            return response, query, references

        if stream:
            for chunk in response:
                chunk_msg = chunk.choices[0].delta.content or ""
                response_message += chunk_msg
                print(chunk_msg, end='')

        else:
            response_message = response.choices[0].message.content

        self.messages.append({'role': 'assistant',
                              'content': response_message})

        if any(references) and print_references:
            ref_msg = ('\n\nThe model was provided with the '
                       'following documents to support its answer:')
            ref_msg += '\n - ' + '\n - '.join(references)
            response_message += ref_msg
            if stream:
                print(ref_msg)

        if debug:
            return response_message, query, references
        else:
            return response_message


class EnergyWizard(EnergyWizardBase):
    """Interface to ask OpenAI LLMs about energy research.

    This class is for execution on a local machine with a vector database in
    memory
    """

    def __init__(self, corpus, model=None, token_budget=3500, ref_col=None):
        """
        Parameters
        ----------
        corpus : pd.DataFrame
            Corpus of text in dataframe format. Must have columns "text" and
            "embedding".
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        token_budget : int
            Number of tokens that can be embedded in the prompt. Note that the
            default budget for GPT-3.5-Turbo is 4096, but you want to subtract
            some tokens to account for the response budget.
        ref_col : None | str
            Optional column label in the corpus that provides a reference text
            string for each chunk of text.
        """

        super().__init__(model, token_budget=token_budget)

        self.corpus = self.preflight_corpus(corpus)
        self.embedding_arr = np.vstack(self.corpus['embedding'].values)
        self.text_arr = self.corpus['text'].values
        self.ref_col = ref_col

    @staticmethod
    def preflight_corpus(corpus, required=('text', 'embedding')):
        """Run preflight checks on the text corpus.

        Parameters
        ----------
        corpus : pd.DataFrame
            Corpus of text in dataframe format. Must have columns "text" and
            "embedding".
        required : list | tuple
            Column names required to be in the corpus df

        Returns
        -------
        corpus : pd.DataFrame
            Corpus of text in dataframe format. Must have columns "text" and
            "embedding".
        """
        missing = [col for col in required if col not in corpus]
        if any(missing):
            msg = ('Text corpus must have {} columns but received '
                   'corpus with columns: {}'
                   .format(missing, list(corpus.columns)))
            raise KeyError(msg)

        if not isinstance(corpus.index.values[0], int):
            corpus['index'] = np.arange(len(corpus))
            corpus = corpus.set_index('index', drop=False)

        return corpus

    def cosine_dist(self, query_embedding):
        """Compute the cosine distance of the query embedding array vs. all of
        the embedding arrays of the full text corpus

        Parameters
        ----------
        query_embedding : np.ndarray
            1D array of the numerical embedding of the request query.

        Returns
        -------
        out : np.ndarray
            1D array with length equal to the number of entries in the text
            corpus. Each value is a distance score where smaller is closer
        """

        dot = np.dot(self.embedding_arr, query_embedding)
        norm1 = np.linalg.norm(query_embedding)
        norm2 = np.linalg.norm(self.embedding_arr, axis=1)

        out = 1 - (dot / (norm1 * norm2))

        return out

    def query_vector_db(self, query, limit=100):
        """Returns a list of strings and relatednesses, sorted from most
        related to least.

        Parameters
        ----------
        query : str
            Question being asked of GPT
        limit : int
            Number of top results to return.

        Returns
        -------
        strings : np.ndarray
            1D array of related strings
        score : np.ndarray
            1D array of float scores of strings
        idx : np.ndarray
            1D array of indices in the text corpus corresponding to the
            ranked strings/scores outputs.
        """

        embedding = self.get_embedding(query)
        scores = 1 - self.cosine_dist(embedding)
        best = np.argsort(scores)[::-1][:limit]

        strings = self.text_arr[best]
        scores = scores[best]

        return strings, scores, best

    def make_ref_list(self, idx):
        """Make a reference list

        Parameters
        ----------
        used_index : np.ndarray
            Indices of the used text from the text corpus

        Returns
        -------
        ref_list : list
            A list of references (strs) used. This takes information straight
            from ``ref_col``. Ideally, this is something like:
            ["{ref_title} ({ref_url})"]
        """
        ref_list = ''
        if self.ref_col is not None and self.ref_col in self.corpus:
            ref_list = list(self.corpus[self.ref_col].iloc[idx].unique())

        return ref_list


class EnergyWizardPostgres(EnergyWizardBase):
    """Interface to ask OpenAI LLMs about energy research.

    This class is for execution with a postgres vector database.
    Querying the database requires the use of the psycopg2 and
    boto3 python packages, environment variables ('EWIZ_DB_USER'
    and 'EWIZ_DB_PASSWORD') storing the db user and password, and
    the specification of other connection paremeters such as host,
    port, and name. The database has the following columns: id,
    embedding, chunks, and metadata.

    This class is designed as follows:
    Vector database: PostgreSQL database accessed using psycopg2.
    Query Embedding: AWS titan using boto3
    LLM Application: GPT4 via Azure deployment
    """
    EMBEDDING_MODEL = 'amazon.titan-embed-text-v1'

    def __init__(self, db_host, db_port, db_name,
                 db_schema, db_table, cursor=None, boto_client=None,
                 model=None, token_budget=3500):
        """
        Parameters
        ----------
        db_host : str
            Host url for postgres database.
        db_port : str
            Port for postres database. ex: '5432'
        db_name : str
            Postgres database name.
        db_schema : str
            Schema name for postres database.
        db_table : str
            Table to query in Postgres database. Necessary columns: id,
            chunks, embedding, title, and url.
        cursor : psycopg2.extensions.cursor
            PostgreSQL database cursor used to execute queries.
        boto_client: botocore.client.BedrockRuntime
            AWS boto3 client used to access embedding model.
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        token_budget : int
            Number of tokens that can be embedded in the prompt. Note that the
            default budget for GPT-3.5-Turbo is 4096, but you want to subtract
            some tokens to account for the response budget.
        """
        boto3 = try_import('boto3')
        psycopg2 = try_import('psycopg2')

        self.db_schema = db_schema
        self.db_table = db_table

        if cursor is None:
            db_user = os.getenv("EWIZ_DB_USER")
            db_password = os.getenv('EWIZ_DB_PASSWORD')
            assert db_user is not None, "Must set EWIZ_DB_USER!"
            assert db_password is not None, "Must set EWIZ_DB_PASSWORD!"
            self.conn = psycopg2.connect(user=db_user,
                                         password=db_password,
                                         host=db_host,
                                         port=db_port,
                                         database=db_name)

            self.cursor = self.conn.cursor()
        else:
            self.cursor = cursor

        if boto_client is None:
            access_key = os.getenv('AWS_ACCESS_KEY_ID')
            secret_key = os.getenv('AWS_SECRET_ACCESS_KEY')
            session_token = os.getenv('AWS_SESSION_TOKEN')
            assert access_key is not None, "Must set AWS_ACCESS_KEY_ID!"
            assert secret_key is not None, ("Must set AWS_SECRET_ACCESS_KEY!")
            assert session_token is not None, "Must set AWS_SESSION_TOKEN!"
            self._aws_client = boto3.client(service_name='bedrock-runtime',
                                            region_name='us-west-2',
                                            aws_access_key_id=access_key,
                                            aws_secret_access_key=secret_key,
                                            aws_session_token=session_token)
        else:
            self._aws_client = boto_client

        super().__init__(model, token_budget=token_budget)

    def get_embedding(self, text):
        """Get the 1D array (list) embedding of a text string
        as generated by AWS Titan.

        Parameters
        ----------
        text : str
            Text to embed

        Returns
        -------
        embedding : list
            List of float that represents the numerical embedding of the text
        """

        body = json.dumps({"inputText": text, })

        model_id = self.EMBEDDING_MODEL
        accept = 'application/json'
        content_type = 'application/json'

        response = self._aws_client.invoke_model(
            body=body,
            modelId=model_id,
            accept=accept,
            contentType=content_type
        )

        response_body = json.loads(response['body'].read())
        embedding = response_body.get('embedding')

        return embedding

    def query_vector_db(self, query, limit=100):
        """Returns a list of strings and relatednesses, sorted from most
        related to least.

        Parameters
        ----------
        query : str
            Question being asked of GPT
        limit : int
            Number of top results to return.

        Returns
        -------
        strings : np.ndarray
            1D array of related strings
        score : np.ndarray
            1D array of float scores of strings
        ids : np.ndarray
            1D array of IDs in the text corpus corresponding to the
            ranked strings/scores outputs.
        """

        query_embedding = self.get_embedding(query)

        self.cursor.execute(f"SELECT {self.db_table}.id, "
                            f"{self.db_table}.chunks, "
                            f"{self.db_table}.embedding "
                            "<=> %s::vector as score "
                            f"FROM {self.db_schema}.{self.db_table} "
                            "ORDER BY embedding <=> %s::vector LIMIT %s;",
                            (query_embedding, query_embedding, limit,), )

        result = self.cursor.fetchall()

        strings = [s[1] for s in result]
        scores = [s[2] for s in result]
        best = [s[0] for s in result]

        return strings, scores, best

    def make_ref_list(self, ids):
        """Make a reference list

        Parameters
        ----------
        used_index : np.ndarray
            IDs of the used text from the text corpus

        Returns
        -------
        ref_list : list
            A list of references (strs) used. Ideally, this is something like:
            ["{ref_title} ({ref_url})"]
        """

        placeholders = ', '.join(['%s'] * len(ids))

        sql_query = (f"SELECT {self.db_table}.title, {self.db_table}.url "
                     f"FROM {self.db_schema}.{self.db_table} "
                     f"WHERE {self.db_table}.id IN (" + placeholders + ")")

        self.cursor.execute(sql_query, ids)

        refs = self.cursor.fetchall()

        ref_strs = (f"{{\"parentTitle\": \"{item[0]}\", "
                    f"\"parentUrl\": \"{item[1]}\"}} " for item in refs)

        unique_values = set(ref_strs)

        ref_list = list(unique_values)

        return ref_list



# -*- coding: utf-8 -*-
"""ELM Web HTML loading with Playwright

We use Playwright so that javascript text is rendered before we scrape.
"""
import logging
from contextlib import AsyncExitStack

from playwright.async_api import async_playwright
from playwright.async_api import Error as PlaywrightError
from playwright.async_api import TimeoutError as PlaywrightTimeoutError

logger = logging.getLogger(__name__)

# block pages by resource type. e.g. image, stylesheet
BLOCK_RESOURCE_TYPES = [
    "beacon",
    "csp_report",
    "font",
    "image",
    "imageset",
    "media",
    "object",
    "texttrack",
    #  can block stylsheets and scripts, though it's not recommended:
    # 'stylesheet',
    # 'script',
    # 'xhr',
]


# block popular 3rd party resources like tracking and advertisements.
BLOCK_RESOURCE_NAMES = [
    "adzerk",
    "analytics",
    "cdn.api.twitter",
    "doubleclick",
    "exelator",
    "facebook",
    "fontawesome",
    "google",
    "google-analytics",
    "googletagmanager",
    "lit.connatix",  # <- not sure about this one
]


async def _intercept_route(route):  # pragma: no cover
    """intercept all requests and abort blocked ones

    Source: https://scrapfly.io/blog/how-to-block-resources-in-playwright/
    """
    if route.request.resource_type in BLOCK_RESOURCE_TYPES:
        return await route.abort()

    if any(key in route.request.url for key in BLOCK_RESOURCE_NAMES):
        return await route.abort()

    return await route.continue_()


async def load_html_with_pw(  # pragma: no cover
    url, browser_semaphore=None, **pw_launch_kwargs
):
    """Extract HTML from URL using Playwright.

    Parameters
    ----------
    url : str
        URL to pull HTML for.
    browser_semaphore : asyncio.Semaphore, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    **pw_launch_kwargs
        Keyword-value argument pairs to pass to
        :meth:`async_playwright.chromium.launch`.

    Returns
    -------
    str
        HTML from page.
    """
    try:
        text = await _load_html(url, browser_semaphore, **pw_launch_kwargs)
    except (PlaywrightError, PlaywrightTimeoutError):
        text = ""
    return text


async def _load_html(  # pragma: no cover
    url, browser_sem=None, **pw_launch_kwargs
):
    """Load html using playwright"""
    if browser_sem is None:
        browser_sem = AsyncExitStack()

    async with async_playwright() as p, browser_sem:
        browser = await p.chromium.launch(**pw_launch_kwargs)
        page = await browser.new_page()
        await page.route("**/*", _intercept_route)
        await page.goto(url)
        await page.wait_for_load_state("networkidle", timeout=90_000)
        text = await page.content()

    return text



# -*- coding: utf-8 -*-
"""ELM Ordinance county file downloading logic"""
import pprint
import asyncio
import logging
from itertools import zip_longest, chain
from contextlib import AsyncExitStack

from elm.ords.llm import StructuredLLMCaller
from elm.ords.extraction import check_for_ordinance_info
from elm.ords.services.threaded import TempFileCache
from elm.ords.validation.location import CountyValidator
from elm.web.document import PDFDocument
from elm.web.file_loader import AsyncFileLoader
from elm.web.google_search import PlaywrightGoogleLinkSearch


logger = logging.getLogger(__name__)
QUESTION_TEMPLATES = [
    '0. "wind energy conversion system zoning ordinances {location}"',
    '1. "{location} wind WECS zoning ordinance"',
    '2. "Where can I find the legal text for commercial wind energy '
    'conversion system zoning ordinances in {location}?"',
    '3. "What is the specific legal information regarding zoning '
    'ordinances for commercial wind energy conversion systems in {location}?"',
]


async def _search_single(
    location, question, browser_sem, num_results=10, **kwargs
):
    """Perform a single google search."""
    if browser_sem is None:
        browser_sem = AsyncExitStack()

    search_engine = PlaywrightGoogleLinkSearch(**kwargs)
    async with browser_sem:
        return await search_engine.results(
            question.format(location=location),
            num_results=num_results,
        )


async def _find_urls(location, num_results=10, browser_sem=None, **kwargs):
    """Parse google search output for URLs."""
    searchers = [
        asyncio.create_task(
            _search_single(
                location, q, browser_sem, num_results=num_results, **kwargs
            ),
            name=location,
        )
        for q in QUESTION_TEMPLATES
    ]
    return await asyncio.gather(*searchers)


def _down_select_urls(search_results, num_urls=5):
    """Select the top 5 URLs."""
    all_urls = chain.from_iterable(
        zip_longest(*[results[0] for results in search_results])
    )
    urls = set()
    for url in all_urls:
        if not url:
            continue
        urls.add(url)
        if len(urls) == num_urls:
            break
    return urls


async def _load_docs(urls, text_splitter, browser_semaphore=None, **kwargs):
    """Load a document for each input URL."""
    loader_kwargs = {
        "html_read_kwargs": {"text_splitter": text_splitter},
        "file_cache_coroutine": TempFileCache.call,
        "browser_semaphore": browser_semaphore,
    }
    loader_kwargs.update(kwargs)
    file_loader = AsyncFileLoader(**loader_kwargs)
    docs = await file_loader.fetch_all(*urls)

    logger.debug(
        "Loaded the following number of pages for docs: %s",
        pprint.PrettyPrinter().pformat(
            {
                doc.metadata.get("source", "Unknown"): len(doc.pages)
                for doc in docs
            }
        ),
    )
    return [doc for doc in docs if not doc.empty]


async def _down_select_docs_correct_location(
    docs, location, county, state, **kwargs
):
    """Remove all documents not pertaining to the location."""
    llm_caller = StructuredLLMCaller(**kwargs)
    county_validator = CountyValidator(llm_caller)
    searchers = [
        asyncio.create_task(
            county_validator.check(doc, county=county, state=state),
            name=location,
        )
        for doc in docs
    ]
    output = await asyncio.gather(*searchers)
    correct_loc_docs = [doc for doc, check in zip(docs, output) if check]
    return sorted(
        correct_loc_docs,
        key=lambda doc: (not isinstance(doc, PDFDocument), len(doc.text)),
    )


async def _check_docs_for_ords(docs, text_splitter, **kwargs):
    """Check documents to see if they contain ordinance info."""
    ord_docs = []
    for doc in docs:
        doc = await check_for_ordinance_info(doc, text_splitter, **kwargs)
        if doc.metadata["contains_ord_info"]:
            ord_docs.append(doc)
    return ord_docs


def _parse_all_ord_docs(all_ord_docs):
    """Parse a list of documents and get the result for the best match."""
    if not all_ord_docs:
        return None

    return sorted(all_ord_docs, key=_ord_doc_sorting_key)[-1]


def _ord_doc_sorting_key(doc):
    """All text sorting key"""
    year, month, day = doc.metadata.get("date", (-1, -1, -1))
    return year, isinstance(doc, PDFDocument), -1 * len(doc.text), month, day


async def download_county_ordinance(
    location,
    text_splitter,
    num_urls=5,
    file_loader_kwargs=None,
    browser_semaphore=None,
    **kwargs
):
    """Download the ordinance document for a single county.

    Parameters
    ----------
    location : elm.ords.utilities.location.Location
        Location objects representing the county.
    text_splitter : obj, optional
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    num_urls : int, optional
        Number of unique Google search result URL's to check for
        ordinance document. By default, ``5``.
    file_loader_kwargs : dict, optional
        Dictionary of keyword-argument pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader` with. The
        "pw_launch_kwargs" key in these will also be used to initialize
        the :class:`elm.web.google_search.PlaywrightGoogleLinkSearch`
        used for the google URL search. By default, ``None``.
    browser_semaphore : asyncio.Semaphore, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument | None
        Document instance for the downloaded document, or ``None`` if no
        document was found.
    """
    file_loader_kwargs = file_loader_kwargs or {}
    pw_launch_kwargs = file_loader_kwargs.get("pw_launch_kwargs", {})
    urls = await _find_urls(
        location.full_name,
        num_results=10,
        browser_sem=browser_semaphore,
        **pw_launch_kwargs
    )
    urls = _down_select_urls(urls, num_urls=num_urls)
    logger.debug("Downloading documents for URLS: \n\t-%s", "\n\t-".join(urls))
    docs = await _load_docs(
        urls, text_splitter, browser_semaphore, **file_loader_kwargs
    )
    docs = await _down_select_docs_correct_location(
        docs,
        location=location.full_name,
        county=location.name,
        state=location.state,
        **kwargs
    )
    docs = await _check_docs_for_ords(docs, text_splitter, **kwargs)
    logger.info(
        "Found %d potential ordinance documents for %s",
        len(docs),
        location.full_name,
    )
    return _parse_all_ord_docs(docs)



# -*- coding: utf-8 -*-
"""ELM Ordinance async decision tree."""
import networkx as nx
import logging

from elm.tree import DecisionTree


logger = logging.getLogger(__name__)


class AsyncDecisionTree(DecisionTree):
    """Async class to traverse a directed graph of LLM prompts. Nodes are
    prompts and edges are transitions between prompts based on conditions
    being met in the LLM response."""

    def __init__(self, graph):
        """Async class to traverse a directed graph of LLM prompts. Nodes are
        prompts and edges are transitions between prompts based on conditions
        being met in the LLM response.

        Parameters
        ----------
        graph : nx.DiGraph
            Directed acyclic graph where nodes are LLM prompts and edges are
            logical transitions based on the response. Must have high-level
            graph attribute "chat_llm_caller" which is a ChatLLMCaller
            instance. Nodes should have attribute "prompt" which can have
            {format} named arguments that will be filled from the high-level
            graph attributes. Edges can have attribute "condition" that is a
            callable to be executed on the LLM response text. An edge from a
            node without a condition acts as an "else" statement if no other
            edge conditions are satisfied. A single edge from node to node
            does not need a condition.
        """
        self._g = graph
        self._history = []
        assert isinstance(self.graph, nx.DiGraph)
        assert "chat_llm_caller" in self.graph.graph

    @property
    def chat_llm_caller(self):
        """elm.ords.llm.ChatLLMCaller: ChatLLMCaller instance for this tree."""
        return self.graph.graph["chat_llm_caller"]

    @property
    def messages(self):
        """Get a list of the conversation messages with the LLM.

        Returns
        -------
        list
        """
        return self.chat_llm_caller.messages

    @property
    def all_messages_txt(self):
        """Get a printout of the full conversation with the LLM

        Returns
        -------
        str
        """
        messages = [
            f"{msg['role'].upper()}: {msg['content']}" for msg in self.messages
        ]
        messages = "\n\n".join(messages)
        return messages

    async def async_call_node(self, node0):
        """Call the LLM with the prompt from the input node and search the
        successor edges for a valid transition condition

        Parameters
        ----------
        node0 : str
            Name of node being executed.

        Returns
        -------
        out : str
            Next node or LLM response if at a leaf node.
        """
        prompt = self._prepare_graph_call(node0)
        out = await self.chat_llm_caller.call(prompt, usage_sub_label="dtree")
        logger.debug(
            "Chat GPT prompt:\n%s\nChat GPT response:\n%s", prompt, out
        )
        return self._parse_graph_output(node0, out)

    async def async_run(self, node0="init"):
        """Traverse the decision tree starting at the input node.

        Parameters
        ----------
        node0 : str
            Name of starting node in the graph. This is typically called "init"

        Returns
        -------
        out : str
            Final response from LLM at the leaf node.
        """

        self._history = []

        while True:
            try:
                out = await self.async_call_node(node0)
            except Exception as e:
                logger.debug(
                    "Error traversing trees, here's the full "
                    "conversation printout:\n%s",
                    self.all_messages_txt,
                )
                last_message = self.messages[-1]["content"]
                msg = (
                    "Ran into an exception when traversing tree. "
                    "Last message from LLM is printed below. "
                    "See debug logs for more detail. "
                    "\nLast message: \n"
                    '"""\n%s\n"""'
                )
                logger.error(msg, last_message)
                logger.exception(e)
                raise RuntimeError(msg % last_message) from e
            if out in self.graph:
                node0 = out
            else:
                break

        logger.info("Output: %s", out)

        return out



# -*- coding: utf-8 -*-
"""ELM Ordinances parsing utilities."""
import json
import logging


logger = logging.getLogger(__name__)


def llm_response_as_json(content):
    """LLM response to JSON.

    Parameters
    ----------
    content : str
        LLM response that contains a string representation of
        a JSON file.

    Returns
    -------
    dict
        Response parsed into dictionary. This dictionary will be empty
        if the response cannot be parsed by JSON.
    """
    content = content.lstrip().rstrip()
    content = content.lstrip("```").lstrip("json").lstrip("\n")
    content = content.rstrip("```")
    content = content.replace("True", "true").replace("False", "false")
    try:
        content = json.loads(content)
    except json.decoder.JSONDecodeError:
        logger.error(
            "LLM returned improperly formatted JSON. "
            "This is likely due to the completion running out of tokens. "
            "Setting a higher token limit may fix this error. "
            "Also ensure you are requesting JSON output in your prompt. "
            "JSON returned:\n%s",
            content,
        )
        content = {}
    return content


# fmt: off
def merge_overlapping_texts(text_chunks, n=300):
    """Merge chunks fo text by removing any overlap.

    Parameters
    ----------
    text_chunks : iterable of str
        Iterable containing text chunks which may or may not contain
        consecutive overlapping portions.
    n : int, optional
        Number of characters to check at the beginning of each message
        for overlap with the previous message. By default, ``100``.

    Returns
    -------
    str
        Merged text.
    """
    if not text_chunks:
        return ""

    out_text = text_chunks[0]
    for next_text in text_chunks[1:]:
        start_ind = out_text[-2 * n:].find(next_text[:n])
        if start_ind == -1:
            out_text = "\n".join([out_text, next_text])
            continue
        start_ind = 2 * n - start_ind
        out_text = "".join([out_text, next_text[start_ind:]])
    return out_text



# -*- coding: utf-8 -*-
# pylint: disable=consider-using-with
"""ELM Ordinance Threaded services"""
import json
import shutil
import asyncio
from pathlib import Path
from functools import partial
from abc import abstractmethod
from tempfile import TemporaryDirectory
from concurrent.futures import ThreadPoolExecutor

from elm.ords.services.base import Service
from elm.web.utilities import write_url_doc_to_file


def _move_file(doc, out_dir):
    """Move a file from a temp directory to an output directory."""
    cached_fp = doc.metadata.get("cache_fn")
    if cached_fp is None:
        return

    cached_fp = Path(cached_fp)
    out_fn = doc.metadata.get("location_name", cached_fp.name)
    if not out_fn.endswith(cached_fp.suffix):
        out_fn = f"{out_fn}{cached_fp.suffix}"

    out_fp = Path(out_dir) / out_fn
    shutil.move(cached_fp, out_fp)
    return out_fp


def _write_cleaned_file(doc, out_dir):
    """Write cleaned ordinance text to directory."""
    cleaned_text = doc.metadata.get("cleaned_ordinance_text")
    location_name = doc.metadata.get("location_name")

    if cleaned_text is None or location_name is None:
        return

    out_fp = Path(out_dir) / f"{location_name} Summary.txt"
    with open(out_fp, "w", encoding="utf-8") as fh:
        fh.write(cleaned_text)
    return out_fp


def _write_ord_db(doc, out_dir):
    """Write parsed ordinance database to directory."""
    ord_db = doc.metadata.get("ordinance_values")
    location_name = doc.metadata.get("location_name")

    if ord_db is None or location_name is None:
        return

    out_fp = Path(out_dir) / f"{location_name} Ordinances.csv"
    ord_db.to_csv(out_fp, index=False)
    return out_fp


_PROCESSING_FUNCTIONS = {
    "move": _move_file,
    "write_clean": _write_cleaned_file,
    "write_db": _write_ord_db,
}


class ThreadedService(Service):
    """Service that contains a ThreadPoolExecutor instance"""

    def __init__(self, **kwargs):
        """

        Parameters
        ----------
        **kwargs
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        self._tpe_kwargs = kwargs or {}
        self.pool = None

    def acquire_resources(self):
        """Open thread pool and temp directory"""
        self.pool = ThreadPoolExecutor(**self._tpe_kwargs)

    def release_resources(self):
        """Shutdown thread pool and cleanup temp directory"""
        self.pool.shutdown(wait=True, cancel_futures=True)


class TempFileCache(ThreadedService):
    """Service that locally caches files downloaded from the internet"""

    def __init__(self, td_kwargs=None, tpe_kwargs=None):
        """

        Parameters
        ----------
        td_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`tempfile.TemporaryDirectory`. By default, ``None``.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self._td_kwargs = td_kwargs or {}
        self._td = None

    @property
    def can_process(self):
        """bool: Always ``True`` (limiting is handled by asyncio)"""
        return True

    def acquire_resources(self):
        """Open thread pool and temp directory"""
        super().acquire_resources()
        self._td = TemporaryDirectory(**self._td_kwargs)

    def release_resources(self):
        """Shutdown thread pool and cleanup temp directory"""
        self._td.cleanup()
        super().release_resources()

    async def process(self, doc, file_content, make_name_unique=False):
        """Write URL doc to file asynchronously.

        Parameters
        ----------
        doc : elm.web.document.Document
            Document containing meta information about the file. Must
            have a "source" key in the `metadata` dict containing the
            URL, which will be converted to a file name using
            :func:`compute_fn_from_url`.
        file_content : str | bytes
            File content, typically string text for HTML files and bytes
            for PDF file.
        make_name_unique : bool, optional
            Option to make file name unique by adding a UUID at the end
            of the file name. By default, ``False``.

        Returns
        -------
        Path
            Path to output file.
        """
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(
            self.pool,
            partial(
                write_url_doc_to_file,
                doc,
                file_content,
                self._td.name,
                make_name_unique=make_name_unique,
            ),
        )
        return result


class StoreFileOnDisk(ThreadedService):
    """Abstract service that manages the storage of a file on disk.

    Storage can occur due to creation or a move of a file.
    """

    def __init__(self, out_dir, tpe_kwargs=None):
        """

        Parameters
        ----------
        out_dir : path-like
            Path to output directory where file should be stored.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self.out_dir = out_dir

    @property
    def can_process(self):
        """bool: Always ``True`` (limiting is handled by asyncio)"""
        return True

    async def process(self, doc):
        """Store file in out directory.

        Parameters
        ----------
        doc : elm.web.document.Document
            Document containing meta information about the file. Must
            have relevant processing keys in the `metadata` dict,
            otherwise the file may not be stored in the output
            directory.

        Returns
        -------
        Path | None
            Path to output file, or `None` if no file was stored.
        """
        return await _run_func_in_pool(
            self.pool,
            partial(_PROCESSING_FUNCTIONS[self._PROCESS], doc, self.out_dir),
        )

    @property
    @abstractmethod
    def _PROCESS(self):
        """str: Key in `_PROCESSING_FUNCTIONS` that defines the doc func."""
        raise NotImplementedError


class FileMover(StoreFileOnDisk):
    """Service that moves files to an output directory"""

    _PROCESS = "move"


class CleanedFileWriter(StoreFileOnDisk):
    """Service that writes cleaned text to a file"""

    _PROCESS = "write_clean"


class OrdDBFileWriter(StoreFileOnDisk):
    """Service that writes cleaned text to a file"""

    _PROCESS = "write_db"


class UsageUpdater(ThreadedService):
    """Service that updates usage info from a tracker into a file."""

    def __init__(self, usage_fp, tpe_kwargs=None):
        """

        Parameters
        ----------
        usage_fp : path-like
            Path to JSON file where usage should be tracked.
        tpe_kwargs : dict, optional
            Keyword-value argument pairs to pass to
            :class:`concurrent.futures.ThreadPoolExecutor`.
            By default, ``None``.
        """
        super().__init__(**(tpe_kwargs or {}))
        self.usage_fp = usage_fp
        self._is_processing = False

    @property
    def can_process(self):
        """bool: ``True`` if file not currently being written to.``"""
        return not self._is_processing

    async def process(self, tracker):
        """Add usage from tracker to file.

        Any existing usage info in the file will remain unchanged
        EXCEPT for anything under the label of the input `tracker`,
        all of which will be replaced with info from the tracker itself.

        Parameters
        ----------
        tracker : elm.ods.services.usage.UsageTracker
            A usage tracker instance that contains usage info to be
            added to output file.
        """
        self._is_processing = True
        await _run_func_in_pool(
            self.pool, partial(_dump_usage, self.usage_fp, tracker)
        )
        self._is_processing = False


async def _run_func_in_pool(pool, callable_fn):
    """Run a callable in process pool"""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(pool, callable_fn)


def _dump_usage(fp, tracker):
    """Dump usage to an existing file."""
    if not Path(fp).exists():
        usage_info = {}
    else:
        with open(fp, "r") as fh:
            usage_info = json.load(fh)

    tracker.add_to(usage_info)
    with open(fp, "w") as fh:
        json.dump(usage_info, fh, indent=4)



# -*- coding: utf-8 -*-
# fmt: off
"""ELM Ordinances CLI."""
import sys
import json
import click
import asyncio
import logging

from elm.version import __version__
from elm.ords.process import process_counties_with_openai


@click.group()
@click.version_option(version=__version__)
@click.pass_context
def main(ctx):
    """ELM ordinances command line interface."""
    ctx.ensure_object(dict)


@main.command()
@click.option("--config", "-c", required=True, type=click.Path(exists=True),
              help="Path to ordinance configuration JSON file. This file "
                   "should contain any/all the arguments to pass to "
                   ":func:`elm.ords.process.process_counties_with_openai`.")
@click.option("-v", "--verbose", is_flag=True,
              help="Flag to show logging on the terminal. Default is not "
                   "to show any logs on the terminal.")
def ords(config, verbose):
    """Download and extract ordinances for a list of counties."""
    with open(config, "r") as fh:
        config = json.load(fh)

    if verbose:
        logger = logging.getLogger("elm")
        logger.addHandler(logging.StreamHandler(stream=sys.stdout))
        logger.setLevel(config.get("log_level", "INFO"))

    # asyncio.run(...) doesn't throw exceptions correctly for some reason...
    loop = asyncio.get_event_loop()
    loop.run_until_complete(process_counties_with_openai(**config))


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    main(obj={})



# -*- coding: utf-8 -*-
"""ELM Ordinance structured parsing class."""
import asyncio
import logging
from copy import deepcopy
from itertools import chain

import pandas as pd

from elm.ords.llm.calling import BaseLLMCaller, ChatLLMCaller
from elm.ords.utilities import llm_response_as_json
from elm.ords.extraction.tree import AsyncDecisionTree
from elm.ords.extraction.features import SetbackFeatures
from elm.ords.extraction.graphs import (
    EXTRACT_ORIGINAL_TEXT_PROMPT,
    setup_graph_wes_types,
    setup_base_graph,
    setup_multiplier,
    setup_conditional,
    setup_participating_owner,
    setup_graph_extra_restriction,
    llm_response_starts_with_yes,
)


logger = logging.getLogger(__name__)

DEFAULT_SYSTEM_MESSAGE = (
    "You are a legal scholar explaining legal ordinances to a wind "
    "energy developer."
)
SETBACKS_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on "
    "ordinances relating to setbacks from {feature} for {wes_type}. Ignore "
    "all text that pertains to private, micro, small, or medium sized wind "
    "energy systems."
)
RESTRICTIONS_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on "
    "ordinances relating to {restriction} for {wes_type}. Ignore "
    "all text that pertains to private, micro, small, or medium sized wind "
    "energy systems."
)
EXTRA_RESTRICTIONS_TO_CHECK = {
    "noise": "maximum noise level",
    "max height": "maximum turbine height",
    "min lot size": "minimum lot size",
    "shadow flicker": "maximum shadow flicker",
    "density": "maximum turbine spacing",
}


def _setup_async_decision_tree(graph_setup_func, **kwargs):
    """Setup Async Decision tree dor ordinance extraction."""
    G = graph_setup_func(**kwargs)
    tree = AsyncDecisionTree(G)
    assert len(tree.chat_llm_caller.messages) == 1
    return tree


def _found_ord(messages):
    """Check if ordinance was found based on messages from the base graph.
    IMPORTANT: This function may break if the base graph structure changes.
    Always update the hardcoded values to match the base graph message
    containing the LLM response about ordinance content.
    """
    if len(messages) < 3:
        return False
    return llm_response_starts_with_yes(messages[2].get("content", ""))


async def _run_async_tree(tree, response_as_json=True):
    """Run Async Decision Tree and return output as dict."""
    try:
        response = await tree.async_run()
    except RuntimeError:
        logger.error(
            "    - NOTE: This is not necessarily an error and may just mean "
            "that the text does not have the requested data."
        )
        response = None

    if response_as_json:
        return llm_response_as_json(response) if response else {}

    return response


async def _run_async_tree_with_bm(tree, base_messages):
    """Run Async Decision Tree from base messages and return dict output."""
    tree.chat_llm_caller.messages = base_messages
    assert len(tree.chat_llm_caller.messages) == len(base_messages)
    return await _run_async_tree(tree)


def _empty_output(feature):
    """Empty output for a feature (not found in text)."""
    if feature in {"struct", "pline"}:
        return [
            {"feature": f"{feature} (participating)"},
            {"feature": f"{feature} (non-participating)"},
        ]
    return [{"feature": feature}]


class StructuredOrdinanceParser(BaseLLMCaller):
    """LLM ordinance document structured data scraping utility."""

    def _init_chat_llm_caller(self, system_message):
        """Initialize a ChatLLMCaller instance for the DecisionTree"""
        return ChatLLMCaller(
            self.llm_service,
            system_message=system_message,
            usage_tracker=self.usage_tracker,
            **self.kwargs,
        )

    async def parse(self, text):
        """Parse text and extract structure ordinance data.

        Parameters
        ----------
        text : str
            Ordinance text which may or may not contain setbacks for one
            or more features (property lines, structure, roads, etc.).
            Text can also contain other supported regulations (noise,
            shadow-flicker, etc,) which will be extracted as well.

        Returns
        -------
        pd.DataFrame
            DataFrame containing parsed-out ordinance values.
        """
        largest_wes_type = await self._check_wind_turbine_type(text)
        logger.info("Largest WES type found in text: %s", largest_wes_type)

        outer_task_name = asyncio.current_task().get_name()
        feature_parsers = [
            asyncio.create_task(
                self._parse_setback_feature(
                    text, feature_kwargs, largest_wes_type
                ),
                name=outer_task_name,
            )
            for feature_kwargs in SetbackFeatures()
        ]
        extras_parsers = [
            asyncio.create_task(
                self._parse_extra_restriction(
                    text, feature, r_text, largest_wes_type
                ),
                name=outer_task_name,
            )
            for feature, r_text in EXTRA_RESTRICTIONS_TO_CHECK.items()
        ]
        outputs = await asyncio.gather(*(feature_parsers + extras_parsers))

        return pd.DataFrame(chain.from_iterable(outputs))

    async def _check_wind_turbine_type(self, text):
        """Get the largest turbine size mentioned in the text."""
        logger.debug("Checking turbine_types")
        tree = _setup_async_decision_tree(
            setup_graph_wes_types,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(DEFAULT_SYSTEM_MESSAGE),
        )
        dtree_wes_types_out = await _run_async_tree(tree)

        largest_wes_type = (
            dtree_wes_types_out.get("largest_wes_type")
            or "large wind energy systems"
        )
        return largest_wes_type

    async def _parse_extra_restriction(
        self, text, feature, restriction_text, largest_wes_type
    ):
        """Parse a non-setback restriction from the text."""
        logger.debug("Parsing extra feature %r", feature)
        system_message = RESTRICTIONS_SYSTEM_MESSAGE.format(
            restriction=restriction_text, wes_type=largest_wes_type
        )
        tree = _setup_async_decision_tree(
            setup_graph_extra_restriction,
            wes_type=largest_wes_type,
            restriction=restriction_text,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
        )
        info = await _run_async_tree(tree)
        info.update({"feature": feature})
        return [info]

    async def _parse_setback_feature(
        self, text, feature_kwargs, largest_wes_type
    ):
        """Parse values for a setback feature."""
        feature = feature_kwargs["feature_id"]
        feature_kwargs["wes_type"] = largest_wes_type
        logger.debug("Parsing feature %r", feature)

        base_messages = await self._base_messages(text, **feature_kwargs)
        if not _found_ord(base_messages):
            logger.debug("Failed `_found_ord` check for feature %r", feature)
            return _empty_output(feature)

        if feature not in {"struct", "pline"}:
            output = {"feature": feature}
            output.update(
                await self._extract_setback_values(
                    text,
                    base_messages=base_messages,
                    **feature_kwargs,
                )
            )
            return [output]

        return await self._extract_setback_values_for_p_or_np(
            text, base_messages, **feature_kwargs
        )

    async def _base_messages(self, text, **feature_kwargs):
        """Get base messages for setback feature parsing."""
        system_message = SETBACKS_SYSTEM_MESSAGE.format(
            feature=feature_kwargs["feature"],
            wes_type=feature_kwargs["wes_type"],
        )
        tree = _setup_async_decision_tree(
            setup_base_graph,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            **feature_kwargs,
        )
        await _run_async_tree(tree, response_as_json=False)
        return deepcopy(tree.chat_llm_caller.messages)

    async def _extract_setback_values_for_p_or_np(
        self, text, base_messages, **feature_kwargs
    ):
        """Extract setback values for participating/non-participating ords."""
        logger.debug("Checking participating vs non-participating")
        dtree_participating_out = await self._run_setback_graph(
            setup_participating_owner,
            text,
            base_messages=deepcopy(base_messages),
            **feature_kwargs,
        )
        outer_task_name = asyncio.current_task().get_name()
        p_or_np_parsers = [
            asyncio.create_task(
                self._parse_p_or_np_text(
                    key, sub_text, base_messages, **feature_kwargs
                ),
                name=outer_task_name,
            )
            for key, sub_text in dtree_participating_out.items()
        ]
        return await asyncio.gather(*p_or_np_parsers)

    async def _parse_p_or_np_text(
        self, key, sub_text, base_messages, **feature_kwargs
    ):
        """Parse participating/non-participating sub-text for ord values."""
        feature = feature_kwargs["feature_id"]
        out_feat_name = f"{feature} ({key})"
        output = {"feature": out_feat_name}
        if not sub_text:
            return output

        feature = feature_kwargs["feature"]
        feature = f"{key} {feature}"
        feature_kwargs["feature"] = feature

        base_messages = deepcopy(base_messages)
        base_messages[-2]["content"] = EXTRACT_ORIGINAL_TEXT_PROMPT.format(
            feature=feature, wes_type=feature_kwargs["wes_type"]
        )
        base_messages[-1]["content"] = sub_text

        values = await self._extract_setback_values(
            sub_text,
            base_messages=base_messages,
            **feature_kwargs,
        )
        output.update(values)
        return output

    async def _extract_setback_values(self, text, **kwargs):
        """Extract setback values for a particular feature from input text."""
        dtree_out = await self._run_setback_graph(
            setup_multiplier, text, **kwargs
        )

        if dtree_out.get("mult_value") is None:
            return dtree_out

        dtree_con_out = await self._run_setback_graph(
            setup_conditional, text, **kwargs
        )
        dtree_out.update(dtree_con_out)
        return dtree_out

    async def _run_setback_graph(
        self,
        graphs_setup_func,
        text,
        feature,
        wes_type,
        base_messages=None,
        **kwargs,
    ):
        """Generic function to run async tree for ordinance extraction."""
        system_message = SETBACKS_SYSTEM_MESSAGE.format(
            feature=feature, wes_type=wes_type
        )
        tree = _setup_async_decision_tree(
            graphs_setup_func,
            feature=feature,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            **kwargs,
        )
        if base_messages:
            return await _run_async_tree_with_bm(tree, base_messages)
        return await _run_async_tree(tree)



# -*- coding: utf-8 -*-
"""ELM Ordinances OpenAI service amd utils."""
import logging

import openai

from elm.base import ApiBase
from elm.ords.services.base import RateLimitedService
from elm.ords.services.usage import TimeBoundedUsageTracker
from elm.utilities.retry import async_retry_with_exponential_backoff


logger = logging.getLogger(__name__)


def usage_from_response(current_usage, response):
    """OpenAI usage parser.

    Parameters
    ----------
    current_usage : dict
        Dictionary containing current usage information. For OpenAI
        trackers, this may contain the keys ``"requests"``,
        ``"prompt_tokens"``, and ``"response_tokens"`` if there is
        already existing tracking information. Empty dictionaries are
        allowed, in which case the three keys above will be added to
        this input.
    response : openai.Completion
        OpenAI Completion object. Must contain a ``usage`` attribute
        that

    Returns
    -------
    dict
        Dictionary with updated usage statistics.
    """
    current_usage["requests"] = current_usage.get("requests", 0) + 1
    current_usage["prompt_tokens"] = (
        current_usage.get("prompt_tokens", 0) + response.usage.prompt_tokens
    )
    current_usage["response_tokens"] = (
        current_usage.get("response_tokens", 0)
        + response.usage.completion_tokens
    )
    return current_usage


def count_tokens(messages, model):
    """Count the number of tokens in an outgoing set of messages.

    Parameters
    ----------
    messages : list
        A list of message objects, where the latter is represented
        using a dictionary. Each message dictionary must have a
        "content" key containing the string to count tokens for.
    model : str
        The OpenAI model being used. This input will be passed to
        :func:`tiktoken.encoding_for_model`.

    Returns
    -------
    int
        Total number of tokens in the set of messages outgoing to
        OpenAI.

    References
    ----------
    https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    """
    message_total = sum(
        ApiBase.count_tokens(message["content"], model=model) + 4
        for message in messages
    )
    return message_total + 3


class OpenAIService(RateLimitedService):
    """OpenAI Chat GPT query service"""

    def __init__(self, client, rate_limit=1e3, rate_tracker=None):
        """

        Parameters
        ----------
        client : openai.AsyncOpenAI | openai.AsyncAzureOpenAI
            Async OpenAI client instance. Must have an async
            `client.chat.completions.create` method.
        rate_limit : int | float, optional
            Token rate limit (typically per minute, but the time
            interval is ultimately controlled by the `rate_tracker`
            instance). By default, ``1e3``.
        rate_tracker : TimeBoundedUsageTracker, optional
            A TimeBoundedUsageTracker instance. This will be used to
            track usage per time interval and compare to `rate_limit`.
            If ``None``, a `TimeBoundedUsageTracker` instance is created
            with default parameters. By default, ``None``.
        """
        super().__init__(rate_limit, rate_tracker or TimeBoundedUsageTracker())
        self.client = client

    async def process(
        self, usage_tracker=None, usage_sub_label="default", *, model, **kwargs
    ):
        """Process a call to OpenAI Chat GPT.

        Note that this method automatically retries queries (with
        backoff) if a rate limit error is throw by the API.

        Parameters
        ----------
        model : str
            OpenAI GPT model to query.
        usage_tracker : `elm.ords.services.usage.UsageTracker`, optional
            UsageTracker instance. Providing this input will update your
            tracker with this call's token usage info.
            By default, ``None``.
        usage_sub_label : str, optional
            Optional label to categorize usage under. This can be used
            to track usage related to certain categories.
            By default, ``"default"``.
        **kwargs
            Keyword arguments to be passed to
            `client.chat.completions.create`.

        Returns
        -------
        str | None
            Chat GPT response as a string, or ``None`` if the call
            failed.
        """
        self._record_prompt_tokens(model, kwargs)
        response = await self._call_gpt(model=model, **kwargs)
        self._record_completion_tokens(response)
        self._record_usage(response, usage_tracker, usage_sub_label)
        return _get_response_message(response)

    def _record_prompt_tokens(self, model, kwargs):
        """Add prompt token count to rate tracker"""
        num_tokens = count_tokens(kwargs.get("messages", []), model)
        self.rate_tracker.add(num_tokens)

    def _record_usage(self, response, usage_tracker, usage_sub_label):
        """Record token usage for user"""
        if usage_tracker is None:
            return
        usage_tracker.update_from_model(response, sub_label=usage_sub_label)

    def _record_completion_tokens(self, response):
        """Add completion token count to rate tracker"""
        if response is None:
            return
        self.rate_tracker.add(response.usage.completion_tokens)

    @async_retry_with_exponential_backoff()
    async def _call_gpt(self, **kwargs):
        """Query Chat GPT with user inputs"""
        try:
            return await self.client.chat.completions.create(**kwargs)
        except openai.BadRequestError as e:
            logger.error("Got 'BadRequestError':")
            logger.exception(e)


def _get_response_message(response):
    """Get message as string from response object"""
    if response is None:
        return None
    return response.choices[0].message.content



# -*- coding: utf-8 -*-
"""ELM Ordinance function to apply ordinance extraction on a document """
import logging
from warnings import warn

from elm.ords.llm import LLMCaller, StructuredLLMCaller
from elm.ords.extraction.date import DateExtractor
from elm.ords.extraction.ordinance import (
    OrdinanceValidator,
    OrdinanceExtractor,
)
from elm.ords.extraction.parse import StructuredOrdinanceParser


logger = logging.getLogger(__name__)


async def check_for_ordinance_info(doc, text_splitter, **kwargs):
    """Parse a single document for ordinance information.

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document potentially containing ordinance information. Note
        that if the document's metadata contains the
        ``"contains_ord_info"`` key, it will not be processed. To force
        a document to be processed by this function, remove that key
        from the documents metadata.
    text_splitter : obj
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance text. The results of
        the parsing are stored in the documents metadata. In particular,
        the metadata will contain a ``"contains_ord_info"`` key that
        will be set to ``True`` if ordinance info was found in the text,
        and ``False`` otherwise. If ``True``, the metadata will also
        contain a ``"date"`` key containing the most recent date that
        the ordinance was enacted (or a tuple of `None` if not found),
        and an ``"ordinance_text"`` key containing the ordinance text
        snippet. Note that the snippet may contain other info as well,
        but should encapsulate all of the ordinance text.
    """
    if "contains_ord_info" in doc.metadata:
        return doc

    llm_caller = StructuredLLMCaller(**kwargs)
    chunks = text_splitter.split_text(doc.text)
    validator = OrdinanceValidator(llm_caller, chunks)
    doc.metadata["contains_ord_info"] = await validator.parse()
    if doc.metadata["contains_ord_info"]:
        doc.metadata["date"] = await DateExtractor(llm_caller).parse(doc)
        doc.metadata["ordinance_text"] = validator.ordinance_text

    return doc


async def extract_ordinance_text_with_llm(doc, text_splitter, extractor):
    """Extract ordinance text from document using LLM.

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document known to contain ordinance information. This means it
        must contain an ``"ordinance_text"`` key in the metadata. You
        can run
        :func:`~elm.ords.extraction.apply.check_for_ordinance_info`
        to have this attribute populated automatically for documents
        that are found to contain ordinance data. Note that if the
        document's metadata does not contain the ``"ordinance_text"``
        key, you will get an error.
    text_splitter : obj
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    extractor : elm.ords.extraction.ordinance.OrdinanceExtractor
        Instance of `~elm.ords.extraction.ordinance.OrdinanceExtractor`
        used for ordinance text extraction.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance text. The results of
        the extraction are stored in the document's metadata. In
        particular, the metadata will contain a
        ``"cleaned_ordinance_text"`` key that will contain the cleaned
        ordinance text.
    """
    text_chunks = text_splitter.split_text(doc.metadata["ordinance_text"])
    ordinance_text = await extractor.check_for_restrictions(text_chunks)
    doc.metadata["restrictions_ordinance_text"] = ordinance_text

    text_chunks = text_splitter.split_text(ordinance_text)
    ordinance_text = await extractor.check_for_correct_size(text_chunks)
    doc.metadata["cleaned_ordinance_text"] = ordinance_text

    return doc


async def extract_ordinance_text_with_ngram_validation(
    doc,
    text_splitter,
    n=4,
    num_extraction_attempts=3,
    ngram_fraction_threshold=0.95,
    **kwargs,
):
    """Extract ordinance text for a single document with known ord info.

    This extraction includes an "ngram" check, which attempts to detect
    wether or not the cleaned text was extracted from the original
    ordinance text. The processing will attempt to re-extract the text
    if the validation does not pass a certain threshold until the
    maximum number of attempts is reached. If the text still does not
    pass validation at this point, there is a good chance that the LLM
    hallucinated parts of the output text, so caution should be taken.

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document known to contain ordinance information. This means it
        must contain an ``"ordinance_text"`` key in the metadata. You
        can run
        :func:`~elm.ords.extraction.apply.check_for_ordinance_info`
        to have this attribute populated automatically for documents
        that are found to contain ordinance data. Note that if the
        document's metadata does not contain the ``"ordinance_text"``
        key, it will not be processed.
    text_splitter : obj
        Instance of an object that implements a `split_text` method.
        The method should take text as input (str) and return a list
        of text chunks. Langchain's text splitters should work for this
        input.
    n : int, optional
        Number of words to include per ngram for the ngram validation,
        which helps ensure that the LLM did not hallucinate.
        By default, ``4``.
    num_extraction_attempts : int, optional
        Number of extraction attempts before returning text that did not
        pass the ngram check. If the processing exceeds this value,
        there is a good chance that the LLM hallucinated parts of the
        output text. Cannot be negative or 0. By default, ``3``.
    ngram_fraction_threshold : float, optional
        Fraction of ngrams in the cleaned text that are also found in
        the original ordinance text for the extraction to be considered
        successful. Should be a value between 0 and 1 (inclusive).
        By default, ``0.95``.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance text. The results of
        the extraction are stored in the document's metadata. In
        particular, the metadata will contain a
        ``"cleaned_ordinance_text"`` key that will contain the cleaned
        ordinance text.
    """
    if not doc.metadata.get("ordinance_text"):
        msg = (
            "Input document has no 'ordinance_text' key or string does not "
            "contain information. Please run `check_for_ordinance_info` "
            "prior to calling this method."
        )
        logger.warning(msg)
        warn(msg, UserWarning)
        return doc

    llm_caller = LLMCaller(**kwargs)
    extractor = OrdinanceExtractor(llm_caller)

    doc = await _extract_with_ngram_check(
        doc,
        text_splitter,
        extractor,
        n=max(1, n),
        num_tries=max(1, num_extraction_attempts),
        ngram_fraction_threshold=max(0, min(1, ngram_fraction_threshold)),
    )

    return doc


async def _extract_with_ngram_check(
    doc,
    text_splitter,
    extractor,
    n=4,
    num_tries=3,
    ngram_fraction_threshold=0.95,
):
    """Extract ordinance info from doc and validate using ngrams."""
    from elm.ords.extraction.ngrams import sentence_ngram_containment

    source = doc.metadata.get("source", "Unknown")
    og_text = doc.metadata["ordinance_text"]
    if not og_text:
        msg = (
            "Document missing original ordinance text! No extraction "
            "performed (Document source: %s)",
            source,
        )
        logger.warning(msg)
        warn(msg, UserWarning)
        return doc

    best_score = 0
    best_summary = ""
    for attempt in range(num_tries):
        doc = await extract_ordinance_text_with_llm(
            doc, text_splitter, extractor
        )
        cleaned_text = doc.metadata["cleaned_ordinance_text"]
        if not cleaned_text:
            logger.debug(
                "No cleaned text found after extraction on attempt %d "
                "for document with source %s. Retrying...",
                attempt,
                source,
            )
            continue

        ngram_frac = sentence_ngram_containment(
            original=og_text, test=cleaned_text, n=n
        )
        if ngram_frac >= ngram_fraction_threshold:
            logger.debug(
                "Document extraction passed ngram check on attempt %d "
                "with score %.2f (Document source: %s)",
                attempt + 1,
                ngram_frac,
                source,
            )
            break

        if ngram_frac > best_score:
            best_score = ngram_frac
            best_summary = cleaned_text

        logger.debug(
            "Document extraction failed ngram check on attempt %d "
            "with score %.2f (Document source: %s). Retrying...",
            attempt + 1,
            ngram_frac,
            source,
        )
    else:
        doc.metadata["cleaned_ordinance_text"] = best_summary
        msg = (
            f"Ngram check failed after {num_tries}. LLM hallucination in "
            "cleaned ordinance text is extremely likely! Proceed with "
            f"caution!! (Document source: {best_score})"
        )
        logger.warning(msg)
        warn(msg, UserWarning)

    return doc


async def extract_ordinance_values(doc, **kwargs):
    """Extract ordinance values for a single document with known ord text.

    Parameters
    ----------
    doc : elm.web.document.BaseDocument
        A document known to contain ordinance text. This means it must
        contain an ``"cleaned_ordinance_text"`` key in the metadata. You
        can run
        :func:`~elm.ords.extraction.apply.extract_ordinance_text_with_llm`
        to have this attribute populated automatically for documents
        that are found to contain ordinance data. Note that if the
        document's metadata does not contain the
        ``"cleaned_ordinance_text"`` key, it will not be processed.
    **kwargs
        Keyword-value pairs used to initialize an
        `elm.ords.llm.LLMCaller` instance.

    Returns
    -------
    elm.web.document.BaseDocument
        Document that has been parsed for ordinance values. The results
        of the extraction are stored in the document's metadata. In
        particular, the metadata will contain an ``"ordinance_values"``
        key that will contain the DataFame with ordinance values.
    """
    if not doc.metadata.get("cleaned_ordinance_text"):
        msg = (
            "Input document has no 'cleaned_ordinance_text' key or string "
            "does not contain info. Please run "
            "`extract_ordinance_text_with_llm` prior to calling this method."
        )
        logger.warning(msg)
        warn(msg, UserWarning)
        return doc

    parser = StructuredOrdinanceParser(**kwargs)
    text = doc.metadata["cleaned_ordinance_text"]
    doc.metadata["ordinance_values"] = await parser.parse(text)
    return doc



"""ELM Ordinance utilities. """

from .counties import load_all_county_info, load_counties_from_fp
from .parsing import llm_response_as_json, merge_overlapping_texts


RTS_SEPARATORS = [
    "Setbacks",
    "CHAPTER ",
    "SECTION ",
    "\r\n\r\n",
    "\r\n",
    "Chapter ",
    "Section ",
    "\n\n",
    "\n",
    "section ",
    "chapter ",
    " ",
    "",
]



# -*- coding: utf-8 -*-
"""ELM Ordinance document content Validation logic

These are primarily used to validate that a legal document applies to a
particular technology (e.g. Large Wind Energy Conversion Systems).
"""
import asyncio
import logging

from elm import ApiBase
from elm.ords.validation.content import (
    ValidationWithMemory,
    possibly_mentions_wind,
)
from elm.ords.utilities.parsing import merge_overlapping_texts


logger = logging.getLogger(__name__)


RESTRICTIONS = """- buildings / structures / residences
- property lines / parcels / subdivisions
- roads / rights-of-way
- railroads
- overhead electrical transmission wires
- bodies of water including wetlands, lakes, reservoirs, streams, and rivers
- natural, wildlife, and environmental conservation areas
- noise restrictions
- shadow flicker restrictions
- density restrictions
- turbine height restrictions
- minimum/maximum lot size
"""


class OrdinanceValidator(ValidationWithMemory):
    """Check document text for wind ordinances."""

    IS_LEGAL_TEXT_PROMPT = (
        "You extract structured data from text. Return your answer in JSON "
        "format (not markdown). Your JSON file must include exactly three "
        "keys. The first key is 'summary', which is a string that provides a "
        "short summary of the text. The second key is 'type', which is a "
        "string that best represent the type of document this text belongs "
        "to. The third key is '{key}', which is a boolean that is set to "
        "True if the type of the text (as you previously determined) is a "
        "legally-binding statute or code and False if the text is an excerpt "
        "from other non-legal text such as a news article, survey, summary, "
        "application, public notice, etc."
    )

    CONTAINS_ORD_PROMPT = (
        "You extract structured data from text. Return your answer in JSON "
        "format (not markdown). Your JSON file must include exactly three "
        "keys. The first key is 'wind_reqs', which is a string that "
        "summarizes the setbacks or other geospatial siting requirements (if "
        "any) given in the text for a wind turbine. The second key is 'reqs', "
        "which lists the quantitative values from the text excerpt that can "
        "be used to compute setbacks or other geospatial siting requirements "
        "for a wind turbine/tower (empty list if none exist in the text). The "
        "last key is '{key}', which is a boolean that is set to True if the "
        "text excerpt provides enough quantitative info to compute setbacks "
        "or other geospatial siting requirements for a wind turbine/tower "
        "and False otherwise. Geospatial siting is impacted by any of the "
        f"following:\n{RESTRICTIONS}"
    )

    IS_UTILITY_SCALE_PROMPT = (
        "You are a legal scholar that reads ordinance text and determines "
        "wether it applies to large wind energy systems. Large wind energy "
        "systems (WES) may also be referred to as wind turbines, wind energy "
        "conversion systems (WECS), wind energy facilities (WEF), wind energy "
        "turbines (WET), large wind energy turbines (LWET), utility-scale "
        "wind energy turbines (UWET), commercial wind energy systems, or "
        "similar. Your client is a commercial wind developer that does not "
        "care about ordinances related to private, micro, small, or medium "
        "sized wind energy systems. Ignore any text related to such systems. "
        "Return your answer in JSON format (not markdown). Your JSON file "
        "must include exactly two keys. The first key is 'summary' which "
        "contains a string that summarizes the types of wind energy systems "
        "the text applies to (if any). The second key is '{key}', which is a "
        "boolean that is set to True if any part of the text excerpt is "
        "applicable to the large wind energy conversion systems that the "
        "client is interested in and False otherwise."
    )

    def __init__(self, structured_llm_caller, text_chunks, num_to_recall=2):
        """

        Parameters
        ----------
        structured_llm_caller : elm.ords.llm.StructuredLLMCaller
            StructuredLLMCaller instance. Used for structured validation
            queries.
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks. This validator may refer to previous text chunks to
            answer validation questions.
        num_to_recall : int, optional
            Number of chunks to check for each validation call. This
            includes the original chunk! For example, if
            `num_to_recall=2`, the validator will first check the chunk
            at the requested index, and then the previous chunk as well.
            By default, ``2``.
        """
        super().__init__(
            structured_llm_caller=structured_llm_caller,
            text_chunks=text_chunks,
            num_to_recall=num_to_recall,
        )
        self._legal_text_mem = []
        self._wind_mention_mem = []
        self._ordinance_chunks = []

    @property
    def is_legal_text(self):
        """bool: ``True`` if text was found to be from a legal source."""
        if not self._legal_text_mem:
            return False
        return sum(self._legal_text_mem) >= 0.5 * len(self._legal_text_mem)

    @property
    def ordinance_text(self):
        """str: Combined ordinance text from the individual chunks."""
        inds_to_grab = set()
        for info in self._ordinance_chunks:
            inds_to_grab |= {
                info["ind"] + x for x in range(1 - self.num_to_recall, 2)
            }

        text = [
            self.text_chunks[ind]
            for ind in sorted(inds_to_grab)
            if 0 <= ind < len(self.text_chunks)
        ]
        return merge_overlapping_texts(text)

    async def parse(self, min_chunks_to_process=3):
        """Parse text chunks and look for ordinance text.

        Parameters
        ----------
        min_chunks_to_process : int, optional
            Minimum number of chunks to process before checking if
            document resembles legal text and ignoring chunks that don't
            pass the wind heuristic. By default, ``3``.

        Returns
        -------
        bool
            ``True`` if any ordinance text was found in the chunks.
        """
        for ind, text in enumerate(self.text_chunks):
            self._wind_mention_mem.append(possibly_mentions_wind(text))
            if ind >= min_chunks_to_process:
                if not self.is_legal_text:
                    return False

                # fmt: off
                if not any(self._wind_mention_mem[-self.num_to_recall:]):
                    continue

            logger.debug("Processing text at ind %d", ind)
            logger.debug("Text:\n%s", text)

            if ind < min_chunks_to_process:
                is_legal_text = await self.parse_from_ind(
                    ind, self.IS_LEGAL_TEXT_PROMPT, key="legal_text"
                )
                self._legal_text_mem.append(is_legal_text)
                if not is_legal_text:
                    logger.debug("Text at ind %d is not legal text", ind)
                    continue

                logger.debug("Text at ind %d is legal text", ind)

            contains_ord_info = await self.parse_from_ind(
                ind, self.CONTAINS_ORD_PROMPT, key="contains_ord_info"
            )
            if not contains_ord_info:
                logger.debug(
                    "Text at ind %d does not contain ordinance info", ind
                )
                continue

            logger.debug("Text at ind %d does contain ordinance info", ind)

            is_utility_scale = await self.parse_from_ind(
                ind, self.IS_UTILITY_SCALE_PROMPT, key="x"
            )
            if not is_utility_scale:
                logger.debug(
                    "Text at ind %d is not for utility-scale WECS", ind
                )
                continue

            logger.debug("Text at ind %d is for utility-scale WECS", ind)

            self._ordinance_chunks.append({"text": text, "ind": ind})
            logger.debug("Added text at ind %d to ordinances", ind)
            # mask, since we got a good result
            self._wind_mention_mem[-1] = False

        return bool(self._ordinance_chunks)


class OrdinanceExtractor:
    """Extract succinct ordinance text from input"""

    SYSTEM_MESSAGE = (
        "You extract one or more direct excerpts from a given text based on "
        "the user's request. Maintain all original formatting and characters "
        "without any paraphrasing. If the relevant text is inside of a "
        "space-delimited table, return the entire table with the original "
        "space-delimited formatting. Never paraphrase! Only return portions "
        "of the original text directly."
    )
    MODEL_INSTRUCTIONS_RESTRICTIONS = (
        "Extract one or more direct text excerpts related to the restrictions "
        "of large wind energy systems with respect to any of the following:\n"
        f"{RESTRICTIONS}"
        "Include section headers (if any) for the text excerpts. Also include "
        "any text excerpts that define what kind of large wind energy "
        "conversion system the restriction applies to. If there is no text "
        "related to siting restrictions of large wind systems, simply say: "
        '"No relevant text."'
    )
    MODEL_INSTRUCTIONS_SIZE = (
        "Extract one or more direct text excerpts pertaining to large wind "
        "energy systems. Large wind energy systems (WES) may also be referred "
        "to as wind turbines, wind energy conversion systems (WECS), wind "
        "energy facilities (WEF), wind energy turbines (WET), large wind "
        "energy turbines (LWET), utility-scale wind energy turbines (UWET), "
        "or similar. Do not return any text excerpts that only apply to "
        "private, micro, small, or medium sized wind energy systems. Include "
        "section headers (if any) for the text excerpts. Also include any "
        "text excerpts that define what kind of large wind energy conversion "
        "system the restriction applies to. If there is no text pertaining to "
        "large wind systems, simply say: "
        '"No relevant text."'
    )

    def __init__(self, llm_caller):
        """

        Parameters
        ----------
        llm_caller : elm.ords.llm.LLMCaller
            LLM Caller instance used to extract ordinance info with.
        """
        self.llm_caller = llm_caller

    async def _process(self, text_chunks, instructions, valid_chunk):
        """Perform extraction processing."""
        logger.info(
            "Extracting ordinance text from %d text chunks asynchronously...",
            len(text_chunks),
        )
        outer_task_name = asyncio.current_task().get_name()
        summaries = [
            asyncio.create_task(
                self.llm_caller.call(
                    sys_msg=self.SYSTEM_MESSAGE,
                    content=f"Text:\n{chunk}\n{instructions}",
                    usage_sub_label="document_ordinance_summary",
                ),
                name=outer_task_name,
            )
            for chunk in text_chunks
        ]
        summary_chunks = await asyncio.gather(*summaries)
        summary_chunks = [
            chunk for chunk in summary_chunks if valid_chunk(chunk)
        ]

        text_summary = "\n".join(summary_chunks)
        logger.debug(
            "Final summary contains %d tokens",
            ApiBase.count_tokens(
                text_summary,
                model=self.llm_caller.kwargs.get("model", "gpt-4"),
            ),
        )
        return text_summary

    async def check_for_restrictions(self, text_chunks):
        """Extract restriction ordinance text from input text chunks.

        Parameters
        ----------
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks.

        Returns
        -------
        str
            Ordinance text extracted from text chunks.
        """
        return await self._process(
            text_chunks=text_chunks,
            instructions=self.MODEL_INSTRUCTIONS_RESTRICTIONS,
            valid_chunk=_valid_chunk_not_short,
        )

    async def check_for_correct_size(self, text_chunks):
        """Extract ordinance text from input text chunks for large WES.

        Parameters
        ----------
        text_chunks : list of str
            List of strings, each of which represent a chunk of text.
            The order of the strings should be the order of the text
            chunks.

        Returns
        -------
        str
            Ordinance text extracted from text chunks.
        """
        return await self._process(
            text_chunks=text_chunks,
            instructions=self.MODEL_INSTRUCTIONS_SIZE,
            valid_chunk=_valid_chunk,
        )


def _valid_chunk(chunk):
    """True if chunk has content."""
    return chunk and "no relevant text" not in chunk.lower()


def _valid_chunk_not_short(chunk):
    """True if chunk has content and is not too short."""
    return _valid_chunk(chunk) and len(chunk) > 20



# -*- coding: utf-8 -*-
"""ELM Web Scraping utilities."""
import uuid
import hashlib
from pathlib import Path

from slugify import slugify


def clean_search_query(query):
    """Check if the first character is a digit and remove it if so.

    Some search tools (e.g., Google) will fail to return results if the
    query has a leading digit: 1. "LangCh..."

    This function will take all the text after the first double quote
    (") if a digit is detected at the beginning of the string.

    Parameters
    ----------
    query : str
        Input query that may or may not contain a leading digit.

    Returns
    -------
    str
        Cleaned query.
    """
    query = query.strip()
    if len(query) < 1:
        return query

    if not query[0].isdigit():
        return query.strip()

    if (first_quote_pos := query[:-1].find('"')) == -1:
        return query.strip()

    last_ind = -1 if query.endswith('"') else None

    # fmt: off
    return query[first_quote_pos + 1:last_ind].strip()


def compute_fn_from_url(url, make_unique=False):
    """Compute a unique file name from URL string.

    File name will always be 128 characters or less, unless the
    `make_unique` argument is set to true. In that case, the max
    length is 164 (a UUID is tagged onto the filename).

    Parameters
    ----------
    url : str
        Input URL to convert into filename.
    make_unique : bool, optional
        Option to add a UUID at the end of the file name to make it
        unique. By default, ``False``.

    Returns
    -------
    str
        Valid filename representation of the URL.
    """
    url = url.replace("https", "").replace("http", "").replace("www", "")
    url = slugify(url)
    url = url.replace("-", "").replace("_", "")

    url = _shorten_using_sha(url)

    if make_unique:
        url = f"{url}{uuid.uuid4()}".replace("-", "")

    return url


def _shorten_using_sha(fn):
    """Reduces FN to 128 characters"""
    if len(fn) <= 128:
        return fn

    out = hashlib.sha256(bytes(fn[64:], encoding="utf-8")).hexdigest()
    return f"{fn[:64]}{out}"


def write_url_doc_to_file(doc, file_content, out_dir, make_name_unique=False):
    """Write a file pulled from URL to disk.

    Parameters
    ----------
    doc : elm.web.document.Document
        Document containing meta information about the file. Must have a
        "source" key in the `metadata` dict containing the URL, which
        will be converted to a file name using
        :func:`compute_fn_from_url`.
    file_content : str | bytes
        File content, typically string text for HTML files and bytes
        for PDF file.
    out_dir : path-like
        Path to directory where file should be stored.
    make_name_unique : bool, optional
        Option to make file name unique by adding a UUID at the end of
        the file name. By default, ``False``.

    Returns
    -------
    Path
        Path to output file.
    """
    out_fn = compute_fn_from_url(
        url=doc.metadata["source"], make_unique=make_name_unique
    )
    out_fp = Path(out_dir) / f"{out_fn}.{doc.FILE_EXTENSION}"
    with open(out_fp, **doc.WRITE_KWARGS) as fh:
        fh.write(file_content)
    return out_fp



# -*- coding: utf-8 -*-
"""ELM Ordinance date extraction logic."""
import logging

logger = logging.getLogger(__name__)


class DateExtractor:
    """Helper class to extract date info from document."""

    SYSTEM_MESSAGE = (
        "You are a legal scholar that reads ordinance text and extracts "
        "structured date information. Return your answer in JSON format (not "
        "markdown). Your JSON file must include exactly four keys. The first "
        "key is 'explanation', which contains a short summary of the most "
        "relevant date information you found in the text. The second key is "
        "'year', which should contain an integer value that represents the "
        "latest year this ordinance was enacted/updated, or null if that "
        "information cannot be found in the text. The third key is 'month', "
        "which should contain an integer value that represents the latest "
        "month of the year this ordinance was enacted/updated, or null if "
        "that information cannot be found in the text. The fourth key is "
        "'day', which should contain an integer value that represents the "
        "latest day of the month this ordinance was enacted/updated, or null "
        "if that information cannot be found in the text."
    )

    def __init__(self, structured_llm_caller):
        """

        Parameters
        ----------
        structured_llm_caller : elm.ords.llm.StructuredLLMCaller
            StructuredLLMCaller instance. Used for structured validation
            queries.
        """
        self.slc = structured_llm_caller

    async def parse(self, doc):
        """Extract date (year, month, day) from doc.

        Parameters
        ----------
        doc : elm.web.document.BaseDocument
            Document with a `raw_pages` attribute.

        Returns
        -------
        tuple
            3-tuple containing year, month, day, or ``None`` if any of
            those are not found.
        """
        all_years = []
        if not doc.raw_pages:
            return None, None, None

        for text in doc.raw_pages:
            if not text:
                continue

            response = await self.slc.call(
                sys_msg=self.SYSTEM_MESSAGE,
                content=f"Please extract the date for this ordinance:\n{text}",
                usage_sub_label="date_extraction",
            )
            if not response:
                continue
            all_years.append(response)

        return _parse_date(all_years)


def _parse_date(json_list):
    """Parse all date elements."""
    year = _parse_date_element(
        json_list,
        key="year",
        max_len=4,
        min_val=2000,
        max_val=float("inf"),
    )
    month = _parse_date_element(
        json_list, key="month", max_len=2, min_val=1, max_val=12
    )
    day = _parse_date_element(
        json_list, key="day", max_len=2, min_val=1, max_val=31
    )

    return year, month, day


def _parse_date_element(json_list, key, max_len, min_val, max_val):
    """Parse out a single date element."""
    date_elements = [info.get(key) for info in json_list]
    logger.debug(f"{key=}, {date_elements=}")
    date_elements = [
        int(y)
        for y in date_elements
        if y is not None
        and len(str(y)) <= max_len
        and (min_val <= int(y) <= max_val)
    ]
    if not date_elements:
        return -1 * float("inf")
    return max(date_elements)



# -*- coding: utf-8 -*-
"""Custom Exceptions and Errors for ELM. """


class ELMError(Exception):
    """Generic ELM Error."""


class ELMRuntimeError(ELMError, RuntimeError):
    """ELM RuntimeError."""
